import collections

from .. import stats
from .. import optim

from . import base


__all__ = ['SGDBaseline']


class SGDBaseline(base.Recommender):
    """Stochastic gradient descent baseline.

    This is somewhat equivalent to `surprise.prediction_algorithms.baseline_only.BaselineOnly` with
    ``method='sgd'``. The difference is that ``surprise`` always uses plain SGD whereas with
    ``creme`` you can use any SGD method, for example `optim.Adam`.

    Parameters:
        row_optimizer (optim.Optimizer): Optimizer used to tune the row weights.
        col_optimizer (optim.Optimizer): Optimizer used to tune the column weights.
        loss (optim.Loss): Loss function to optimize for.
        l2 (float): regularization amount used to push weights towards 0.
        global_mean (stats.Mean)

    Attributes:
        row_biases (collections.defaultdict)
        col_biases (collections.defaultdict)

    Example:

        ::

            >>> from creme import reco

            >>> ratings = [
            ...     ('Alice', 'Superman', 8),
            ...     ('Alice', 'Terminator', 9),
            ...     ('Alice', 'Star Wars', 8),
            ...     ('Alice', 'Notting Hill', 2),
            ...     ('Alice', 'Harry Potter ', 5),
            ...     ('Bob', 'Superman', 8),
            ...     ('Bob', 'Terminator', 9),
            ...     ('Bob', 'Star Wars', 8),
            ...     ('Bob', 'Notting Hill', 2),
            ... ]

            >>> model = reco.SGDBaseline(
            ...     row_optimizer=optim.SGD(0.005),
            ...     col_optimizer=optim.SGD(0.005)
            ... )

            >>> for user, movie, rating in ratings:
            ...     _ = model.fit_one(user, movie, rating)

            >>> model.predict_one('Bob', 'Harry Potter')
            6.552432...

    """

    def __init__(self, row_optimizer=None, col_optimizer=None, loss=None, l2=0.):
        super().__init__()
        self.row_optimizer = optim.SGD() if row_optimizer is None else row_optimizer
        self.col_optimizer = optim.SGD() if col_optimizer is None else row_optimizer
        self.loss = optim.losses.Squared() if loss is None else row_optimizer
        self.l2 = l2
        self.global_mean = stats.Mean()
        self.row_biases = collections.defaultdict(float)
        self.col_biases = collections.defaultdict(float)

    def fit_one(self, r_id, c_id, y):

        self.row_optimizer.update_before_pred(w=self.row_biases)
        self.col_optimizer.update_before_pred(w=self.col_biases)

        # Predict the value
        y_pred = self.predict_one(r_id, c_id)

        # Compute the gradient of the loss with respect to the prediction
        loss_gradient = self.loss.gradient(y, y_pred)

        # Update the row biases
        r_grad = {r_id: loss_gradient + self.l2 * self.row_biases.get(r_id, 0)}
        self.row_biases = self.row_optimizer.update_after_pred(w=self.row_biases, g=r_grad)

        # Update the row biases
        c_grad = {c_id: loss_gradient + self.l2 * self.col_biases.get(c_id, 0)}
        self.col_biases = self.col_optimizer.update_after_pred(w=self.col_biases, g=c_grad)

        # Update the global mean
        self.global_mean.update(y)

        return y_pred

    def predict_one(self, r_id, c_id):
        return (self.global_mean.get() or 0) + self.row_biases.get(r_id, 0) + self.col_biases.get(c_id, 0)



from . import base
from . import covariance
from . import var


class PearsonCorrelation(base.Bivariate):
    """Online Pearson correlation.

    Parameters:
        ddof (int): Delta Degrees of Freedom. Defaults to ``1``.

    Attributes:
        var_x (stats.Var): Running variance of `x`.
        var_y (stats.Var): Running variance of `y`.
        cov_xy (stats.Var): Running covariance of `x` and `y`.

    Example:

        ::

            >>> from creme import stats

            >>> x = [0, 0, 0, 1, 1, 1, 1]
            >>> y = [0, 1, 2, 3, 4, 5, 6]

            >>> pearson = stats.PearsonCorrelation()

            >>> for xi, yi in zip(x, y):
            ...     print(pearson.update(xi, yi).get())
            0
            0
            0
            0.774596...
            0.866025...
            0.878310...
            0.866025...

    """

    def __init__(self, ddof=1):
        self.var_x = var.Var(ddof=ddof)
        self.var_y = var.Var(ddof=ddof)
        self.cov_xy = covariance.Covariance(ddof=ddof)

    def update(self, x, y):
        self.var_x.update(x)
        self.var_y.update(y)
        self.cov_xy.update(x, y)
        return self

    def get(self):
        var_x = self.var_x.get()
        var_y = self.var_y.get()
        if var_x and var_y:
            return self.cov_xy.get() / (var_x * var_y) ** 0.5
        return 0



import collections

from . import base


__all__ = ['AdaGrad']


class AdaGrad(base.Optimizer):
    """AdaGrad optimizer.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_breast_cancer(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> optimizer = optim.AdaGrad()
            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LogisticRegression(optimizer)
            ... )
            >>> metric = metrics.F1()

            >>> model_selection.online_score(X_y, model, metric)
            F1: 0.970547

    References:
        1. `Adaptive subgradient methods for online learning andstochastic optimization <http://www.jmlr.org/papers/volume12/duchi11a/duchi11a.pdf>`_

    """

    def __init__(self, lr=0.1, eps=1e-8):
        super().__init__(lr)
        self.eps = eps
        self.g2 = collections.defaultdict(float)

    def _update_after_pred(self, w, g):

        for i, gi in g.items():
            self.g2[i] += gi ** 2
            w[i] -= self.learning_rate / (self.g2[i] + self.eps) ** 0.5 * gi

        return w



import collections
import random

from .. import base
from .. import preprocessing


Limits = collections.namedtuple('Limits', 'lower upper')

Split = collections.namedtuple('Split', 'feature value')


class Node:

    def __init__(self, depth, size_limit, split=None, left=None, right=None):
        self.split = split
        self.size_limit = size_limit
        self.depth = depth
        self.left = left
        self.right = right
        self.r_mass = 0
        self.l_mass = 0

    @property
    def is_terminal(self):
        return self.left is None

    def next(self, x):
        if x[self.split.feature] < self.split.value:
            return self.left
        return self.right

    def update(self, x):

        # Increment the mass
        self.l_mass += 1

        # Update the next node
        if not self.is_terminal:
            self.next(x).update(x)

    def score(self, x):
        if self.r_mass < self.size_limit or self.is_terminal:
            return self.r_mass * 2 ** self.depth
        return self.next(x).score(x)

    def pivot_masses(self):
        self.r_mass = self.l_mass
        self.l_mass = 0

        if not self.is_terminal:
            self.left.pivot_masses()
            self.right.pivot_masses()

    def __str__(self):
        s = '\t' * self.depth + f'r_mass={self.r_mass}, l_mass={self.l_mass}'

        if self.is_terminal:
            return s

        return f'{s} ({self.split.feature} < {self.split.value:.3f})\n{self.left}\n{self.right}'


def make_limits(rng):
    sq = rng.random()
    return Limits(lower=sq - 2 * max(sq, 1 - sq), upper=sq + 2 * max(sq, 1 - sq))


def make_tree(limits, tree_height, depth, size_limit, rng):
    """Returns a half-space tree with splits occurring between 0 and 1.

    Parameters:
        limits (dict)
        max_depth (int): The desired tree height.
        depth (int)
        size_limit (int)

    """

    if depth == tree_height:
        return Node(depth=depth, size_limit=size_limit)

    # Randomly pick a feature and find the center of it's limits
    feature = rng.choice(list(limits.keys()))
    center = (limits[feature].lower + limits[feature].upper) / 2

    # Build the left node
    tmp = limits[feature]
    limits[feature] = Limits(lower=tmp.lower, upper=center)
    left = make_tree(
        limits=limits,
        tree_height=tree_height,
        depth=depth + 1,
        size_limit=size_limit,
        rng=rng
    )
    limits[feature] = tmp

    # Build the right node
    tmp = limits[feature]
    limits[feature] = Limits(lower=center, upper=tmp.upper)
    right = make_tree(
        limits=limits,
        tree_height=tree_height,
        depth=depth + 1,
        size_limit=size_limit,
        rng=rng
    )
    limits[feature] = tmp

    return Node(
        depth=depth,
        split=Split(feature=feature, value=center),
        left=left,
        right=right,
        size_limit=size_limit
    )


class HalfSpaceTrees(base.OutlierDetector):
    """Half-Space Trees (HST).

    Half-space trees are an online variant of isolation forests. They work well when anomalies are
    spread out in time. They do not work well if anomalies are packed together in windows of time,
    such as in the KDD'99 HTTP and SMTP datasets.

    Parameters:
        n_trees (int): Number of trees to use.
        tree_height (int): Height of each tree.
        window_size (int): Number of observations to use for calculating the mass at each node in
            each tree.
        scale (bool): Whether or not to scale features between 0 and 1. Only set to ``False`` if
            you know that all your features are already contained between 0 and 1.
        seed (int): Random number seed.

    Example:

        ::

            >>> from creme import anomaly

            >>> X = [0.5, 0.45, 0.43, 0.44, 0.445, 0.45, 0.0]
            >>> hst = anomaly.HalfSpaceTrees(
            ...     n_trees=5,
            ...     tree_height=3,
            ...     window_size=3,
            ...     scale=False,
            ...     seed=42
            ... )

            >>> for x in X[:3]:
            ...     hst = hst.fit_one({'x': x})  # Warming up

            >>> for x in X:
            ...     features = {'x': x}
            ...     hst = hst.fit_one(features)
            ...     print(f'Anomaly score for {x:.3f}: {hst.score_one(features)}')
            Anomaly score for 0.500: 120
            Anomaly score for 0.450: 120
            Anomaly score for 0.430: 120
            Anomaly score for 0.440: 120
            Anomaly score for 0.445: 120
            Anomaly score for 0.450: 120
            Anomaly score for 0.000: 0

    References:
        1. `Fast Anomaly Detection for Streaming Data <https://www.ijcai.org/Proceedings/11/Papers/254.pdf>`_

    """

    def __init__(self, n_trees=25, tree_height=15, window_size=250, scale=True, seed=None):
        self.n_trees = n_trees
        self.window_size = window_size
        self.tree_height = tree_height
        self.scale = scale
        self.rng = random.Random(seed)
        self.trees = []
        self.min_max_scaler = preprocessing.MinMaxScaler()
        self.counter = 0

    def fit_one(self, x):

        # Scale the features between 0 and 1
        if self.scale:
            x = self.min_max_scaler.transform_one(x)
            self.min_max_scaler.fit_one(x)

        # The trees are built when the first observation comes in
        if not self.trees:
            self.trees = [
                make_tree(
                    limits={f: make_limits(rng=self.rng) for f in x},
                    tree_height=self.tree_height,
                    depth=0,
                    size_limit=.1 * self.window_size,
                    rng=self.rng
                )
                for _ in range(self.n_trees)
            ]

        # Update each tree
        for tree in self.trees:
            tree.update(x)

        # Pivot the masses if necessary
        self.counter += 1
        if self.counter == self.window_size:
            for tree in self.trees:
                tree.pivot_masses()
            self.counter = 0

        return self

    def score_one(self, x):
        if not self.trees:
            return 0
        return sum(tree.score(x) for tree in self.trees)



import collections

from .. import stats

from . import base
from . import precision


__all__ = [
    'MacroRecall',
    'MicroRecall',
    'Recall',
    'WeightedRecall'
]


class BaseRecall:

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True


class Recall(stats.Mean, BaseRecall, base.BinaryMetric):
    """Binary recall score.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [True, False, True, True, True]
            >>> y_pred = [True, True, False, True, True]

            >>> metric = metrics.Recall()

            >>> for yt, yp in zip(y_true, y_pred):
            ...     print(metric.update(yt, yp))
            Recall: 1.
            Recall: 1.
            Recall: 0.5
            Recall: 0.666667
            Recall: 0.75

    """

    def update(self, y_true, y_pred, sample_weight=1.):
        if y_true:
            return super().update(x=y_true == y_pred, w=sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        if y_true:
            return super().revert(x=y_true == y_pred, w=sample_weight)
        return self


class MacroRecall(BaseRecall, base.MultiClassMetric):
    """Macro-average recall score.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [0, 1, 2, 2, 2]
            >>> y_pred = [0, 0, 2, 2, 1]

            >>> metric = metrics.MacroRecall()

            >>> for yt, yp in zip(y_true, y_pred):
            ...     print(metric.update(yt, yp))
            MacroRecall: 1.
            MacroRecall: 0.5
            MacroRecall: 0.666667
            MacroRecall: 0.666667
            MacroRecall: 0.555556

    """

    def __init__(self):
        self.recalls = collections.defaultdict(Recall)
        self._class_counts = collections.Counter()

    def update(self, y_true, y_pred, sample_weight=1.):
        self.recalls[y_true].update(True, y_true == y_pred, sample_weight)
        self._class_counts.update([y_true, y_pred])
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.recalls[y_true].revert(True, y_true == y_pred, sample_weight)
        self._class_counts.subtract([y_true, y_pred])
        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return sum(self.recalls[c].get() for c in relevant) / len(relevant)
        except ZeroDivisionError:
            return 0.


class MicroRecall(precision.MicroPrecision):
    """Micro-average recall score.

    The micro-average recall is exactly equivalent to the micro-average precision as well as the
    micro-average F1 score.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [0, 1, 2, 2, 2]
            >>> y_pred = [0, 0, 2, 2, 1]

            >>> metric = metrics.MicroRecall()

            >>> for yt, yp in zip(y_true, y_pred):
            ...     print(metric.update(yt, yp))
            MicroRecall: 1.
            MicroRecall: 0.5
            MicroRecall: 0.666667
            MicroRecall: 0.75
            MicroRecall: 0.6

    References:
        1. `Why are precision, recall and F1 score equal when using micro averaging in a multi-class problem? <https://simonhessner.de/why-are-precision-recall-and-f1-score-equal-when-using-micro-averaging-in-a-multi-class-problem/>`_

    """


class WeightedRecall(BaseRecall, base.MultiClassMetric):
    """Weighted-average recall score.

    This uses the support of each label to compute an average score, whereas `MacroRecall` ignores
    the support.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [0, 1, 2, 2, 2]
            >>> y_pred = [0, 0, 2, 2, 1]

            >>> metric = metrics.WeightedRecall()

            >>> for yt, yp in zip(y_true, y_pred):
            ...     print(metric.update(yt, yp))
            WeightedRecall: 1.
            WeightedRecall: 0.5
            WeightedRecall: 0.666667
            WeightedRecall: 0.75
            WeightedRecall: 0.6

    """

    def __init__(self):
        self.recalls = collections.defaultdict(Recall)
        self.support = collections.Counter()
        self._class_counts = collections.Counter()

    def update(self, y_true, y_pred, sample_weight=1.):
        self.recalls[y_true].update(True, y_true == y_pred, sample_weight)
        self.support.update({y_true: sample_weight})
        self._class_counts.update([y_true, y_pred])
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.recalls[y_true].revert(True, y_true == y_pred, sample_weight)
        self.support.subtract({y_true: sample_weight})
        self._class_counts.subtract([y_true, y_pred])
        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return (
                sum(self.recalls[c].get() * self.support[c] for c in relevant) /
                sum(self.support[c] for c in relevant)
            )
        except ZeroDivisionError:
            return 0.



"""Feature extraction from a stream."""
from .agg import Agg
from .agg import TargetAgg
from .differ import Differ
from .vectorize import CountVectorizer
from .vectorize import TFIDFVectorizer


__all__ = [
    'Agg',
    'CountVectorizer',
    'Differ',
    'TargetAgg',
    'TFIDFVectorizer'
]



from . import base


__all__ = ['SGD']


class SGD(base.Optimizer):
    """Plain stochastic gradient descent.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_breast_cancer(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> optimizer = optim.SGD(0.1)
            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LogisticRegression(optimizer)
            ... )
            >>> metric = metrics.F1()

            >>> model_selection.online_score(X_y, model, metric)
            F1: 0.964739

    References:
        1. `A Stochastic Approximation Method <https://pdfs.semanticscholar.org/34dd/d8865569c2c32dec9bf7ffc817ff42faaa01.pdf>`_

    """

    def __init__(self, lr=0.01):
        super().__init__(lr)

    def _update_after_pred(self, w, g):

        for i, gi in g.items():
            w[i] -= self.learning_rate * gi

        return w



import collections
import copy
import functools

from .. import base
from .. import optim
from .. import utils


__all__ = ['SoftmaxRegression']


class SoftmaxRegression(base.MultiClassifier):
    """Softmax regression is a generalization of logistic regression to multiple classes.

    Softmax regression is also known as "multinomial logistic regression". There are a set weights
    for each class, hence the ``weights`` attribute is a nested `collections.defaultdict`. The main
    advantage of using this instead of a one-vs-all logistic regression is that the probabilities
    will be calibrated. Moreover softmax regression is more robust to outliers.

    Parameters:
        optimizer (optim.Optimizer): The sequential optimizer used to find the best weights.
        loss (optim.MultiClassificationLoss): The loss function to optimize for.
        l2 (float): Amount of L2 regularization used to push weights towards 0.

    Attributes:
        weights (collections.defaultdict)

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_iris(),
            ...     shuffle=True,
            ...     random_state=42
            ... )

            >>> model = preprocessing.PolynomialExtender(degree=2)
            >>> model |= preprocessing.StandardScaler()
            >>> model |= linear_model.SoftmaxRegression()

            >>> metric = metrics.MacroF1()

            >>> model_selection.online_score(X_y, model, metric)
            MacroF1: 0.809381

    References:

        1. `Stochastic gradient descent; Classification <https://www.inf.ed.ac.uk/teaching/courses/mlp/2016/mlp02-sln.pdf>`_
        2. `Binary vs. Multi-Class Logistic Regression <https://chrisyeh96.github.io/2018/06/11/logistic-regression.html>`_

    """

    def __init__(self, optimizer=None, loss=None, l2=0):
        self.optimizers = collections.defaultdict(functools.partial(
            copy.deepcopy,
            optim.SGD(0.01) if optimizer is None else optimizer
        ))
        self.loss = optim.losses.CrossEntropy() if loss is None else loss
        self.l2 = l2
        self.weights = collections.defaultdict(functools.partial(collections.defaultdict, float))

    def fit_one(self, x, y):

        # Some optimizers need to do something before a prediction is made
        for label, weights in self.weights.items():
            self.optimizers[label].update_before_pred(w=weights)

        # Make a prediction for the given features
        y_pred = self.predict_proba_one(x)

        # Compute the gradient of the loss w.r.t. each label
        loss_gradients = self.loss.gradient(y_true=y, y_pred=y_pred)

        for label, loss in loss_gradients.items():

            # Compute the gradient w.r.t. each feature
            weights = self.weights[label]
            gradient = {i: xi * loss + self.l2 * weights.get(i, 0) for i, xi in x.items()}
            self.weights[label] = self.optimizers[label].update_after_pred(w=weights, g=gradient)

        return self

    def predict_proba_one(self, x):
        return utils.math.softmax({
            label: utils.math.dot(weights, x)
            for label, weights in self.weights.items()
        })



from .. import base


def always_true(x):
    return True


class Differ(base.Transformer):
    """Calculates value differences between different observations.

    Parameters:
        on (str): The name of the feature for which to compute differences. The type of the values
            must support the ``-`` operator, for example an `int` or a `datetime.timedelta`.
        when (callable): A function which indicates when an event occurs. This can used to compute
            the difference since the last time an event occured. If ``True`` then the differences
            will always be computed with regards to the last observation.
        by (str): Can be used to compute value differences for different groups. If ``None`` then
            the differences will be computed globally.
        when_name (str): The name of the of the feature that is returned is automatically built
            from the input parameters. If ``when`` is a ``lambda`` function then no name can be
            inferred. You can thus use this parameter to indicate the name of the event.

    Attributes:
        last_moments (dict): Indicates the last value for each value in the ``by`` field.
        feature_name (str): The name of the resulting feature.

    Example:

        ::

            >>> import datetime as dt

            >>> data = [
            ...     {'weather': 'sunny', 'moment': 1, 'country': 'Sweden'},
            ...     {'weather': 'rainy', 'moment': 1, 'country': 'Rwanda'},
            ...     {'weather': 'rainy', 'moment': 2, 'country': 'Rwanda'},
            ...     {'weather': 'rainy', 'moment': 2, 'country': 'Sweden'},
            ...     {'weather': 'sunny', 'moment': 3, 'country': 'Rwanda'},
            ...     {'weather': 'rainy', 'moment': 4, 'country': 'Rwanda'},
            ...     {'weather': 'rainy', 'moment': 3, 'country': 'Sweden'},
            ...     {'weather': 'sunny', 'moment': 4, 'country': 'Sweden'}
            ... ]

            >>> def is_sunny(x):
            ...     return x['weather'] == 'sunny'

            >>> differ = Differ(
            ...     on='moment',
            ...     by='country',
            ...     when=is_sunny
            ... )

            >>> for x in data:
            ...     differ = differ.fit_one(x)
            ...     print(differ.transform_one(x))
            {'moment_diff_since_is_sunny_by_country': 0}
            {}
            {}
            {'moment_diff_since_is_sunny_by_country': 1}
            {'moment_diff_since_is_sunny_by_country': 0}
            {'moment_diff_since_is_sunny_by_country': 1}
            {'moment_diff_since_is_sunny_by_country': 2}
            {'moment_diff_since_is_sunny_by_country': 0}

    """

    def __init__(self, on, when=True, by=None, when_name=None):
        self.on = on
        self.by = by
        self.when = always_true if when is True else when
        self.when_name = when.__name__ if when_name is None and when is not True else when_name
        self.last_moments = {}
        self.feature_name = f'{self.on}_diff'
        if when is not True:
            self.feature_name += f'_since_{self.when_name}'
        if by is not None:
            self.feature_name += f'_by_{self.by}'

    def fit_one(self, x, y=None):

        if self.when(x):
            self.last_moments[x[self.by]] = x[self.on]

        return self

    def transform_one(self, x):
        if self.when(x):
            return {self.feature_name: x[self.on] - x[self.on]}
        if x[self.by] in self.last_moments:
            return {self.feature_name: x[self.on] - self.last_moments[x[self.by]]}
        return {}



import abc
import math

from .. import base
from .. import optim


__all__ = ['HedgeRegressor']


class Hedge(base.Ensemble):

    def __init__(self, models, weights, learning_rate):
        super().__init__(models)
        self.weights = [1] * len(models)
        self.learning_rate = learning_rate

    @abc.abstractmethod
    def _get_loss(self, model, x, y):
        """Returns a prediction."""

    def fit_one(self, x, y):

        # Make a prediction and update the weights accordingly for each model
        total = 0
        for i, model in enumerate(self):
            loss = self._get_loss(model=model, x=x, y=y)
            self.weights[i] *= math.exp(-self.learning_rate * loss)
            total += self.weights[i]
            model.fit_one(x, y)

        # Normalize the weights so that they sum up to 1
        if total:
            for i, _ in enumerate(self.weights):
                self.weights[i] /= total

        return self


class HedgeRegressor(Hedge, base.Regressor):
    """Hedge Algorithm for regression.

    The Hedge Algorithm is a special case of the Weighted Majority Algorithm for arbitrary losses.

    Parameters:
        regressors (list of `base.Regressor`): The set of regressor to hedge.
        weights (list of `float`): The initial weight of each model. If ``None`` then a uniform set
            of weights is assumed. This roughly translates to the prior amount of trust we have in
            each model.
        loss (optim.RegressionLoss): The loss function that has to be minimized. Defaults to
            `optim.losses.Squared`.
        learning_rate (float): The learning rate by which the model weights are multiplied at each
            iteration.

    Example:

        ::

            >>> from creme import ensemble
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> optimizers = [
            ...     optim.SGD(0.01),
            ...     optim.RMSProp(),
            ...     optim.AdaGrad()
            ... ]

            >>> for optimizer in optimizers:
            ...
            ...     X_y = stream.iter_sklearn_dataset(
            ...         dataset=datasets.load_boston(),
            ...         shuffle=False
            ...     )
            ...     metric = metrics.MAE()
            ...     model = (
            ...         preprocessing.StandardScaler() |
            ...         linear_model.LinearRegression(
            ...             optimizer=optimizer,
            ...             intercept_lr=.1
            ...         )
            ...     )
            ...
            ...     print(optimizer, model_selection.online_score(X_y, model, metric))
            SGD MAE: 7.203527
            RMSProp MAE: 3.312368
            AdaGrad MAE: 3.984558

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_boston(),
            ...     shuffle=False
            ... )
            >>> metric = metrics.MAE()
            >>> hedge = (
            ...     preprocessing.StandardScaler() |
            ...     ensemble.HedgeRegressor(
            ...         regressors=[
            ...             linear_model.LinearRegression(optimizer=o, intercept_lr=.1)
            ...             for o in optimizers
            ...         ]
            ...     )
            ... )

            >>> model_selection.online_score(X_y, hedge, metric)
            MAE: 3.245318

    References:
        1. `Online Learning from Experts: Weighed Majority and Hedge <https://www.shivani-agarwal.net/Teaching/E0370/Aug-2011/Lectures/20-scribe1.pdf>`_
        2. `Multiplicative weight update method <https://www.wikiwand.com/en/Multiplicative_weight_update_method>`_
        3. `Exponentiated Gradient versus GradientDescent for Linear Predictors <https://users.soe.ucsc.edu/~manfred/pubs/J36.pdf>`_

    """

    def __init__(self, regressors, weights=None, loss=None, learning_rate=0.5):
        super().__init__(
            models=regressors,
            weights=weights,
            learning_rate=learning_rate
        )
        self.loss = optim.losses.Squared() if loss is None else loss

    def _get_loss(self, model, x, y):
        return self.loss.eval(y_true=y, y_pred=model.predict_one(x))

    def predict_one(self, x):
        return sum(model.predict_one(x) * weight for model, weight in zip(self, self.weights))



cimport base
cimport mean

from . import base
from . import mean


cdef class Var(base.Univariate):
    """Running variance using Welford's algorithm.

    Parameters:
        ddof (int): Delta Degrees of Freedom. The divisor used in calculations is $n$ - ddof,
            where $n$ represents the number of seen elements.

    Attributes:
        mean (stats.Mean): The running mean.
        sos (float): The running sum of squares.

    Example:

        ::

            >>> import creme.stats

            >>> X = [3, 5, 4, 7, 10, 12]

            >>> var = creme.stats.Var()
            >>> for x in X:
            ...     print(var.update(x).get())
            0.0
            2.0
            1.0
            2.916666...
            7.7
            12.56666...

    References:

        1. `Welford's online algorithm <https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Welford's_Online_algorithm>`_

    """

    cdef readonly long ddof
    cdef readonly mean.Mean mean
    cdef readonly double sos

    def __init__(self, ddof=1):
        self.ddof = ddof
        self.mean = mean.Mean()
        self.sos = 0.

    cpdef Var update(self, double x):
        mean = self.mean.get()
        self.mean.update(x)
        self.sos += (x - mean) * (x - self.mean.get())
        return self

    cpdef double get(self):
        if self.mean.n > self.ddof:
            return self.sos / (self.mean.n - self.ddof)
        return 0.


class RollingVar(base.RollingUnivariate):
    """Running variance over a window.

    Parameters:
        window_size (int): Size of the rolling window.
        ddof (int): Delta Degrees of Freedom for the variance.

    Attributes:
        sos (float): Sum of squares.
        rolling_mean (stats.RollingMean): The running rolling mean.

    Example:

        ::

            >>> import creme

            >>> X = [1, 4, 2, -4, -8, 0]

            >>> rolling_variance = creme.stats.RollingVar(ddof=1, window_size=2)
            >>> for x in X:
            ...     print(rolling_variance.update(x).get())
            0.0
            4.5
            2.0
            18.0
            8.0
            32.0

            >>> rolling_variance = creme.stats.RollingVar(ddof=1, window_size=3)
            >>> for x in X:
            ...     print(rolling_variance.update(x).get())
            0.0
            4.5
            2.333333...
            17.333333...
            25.333333...
            16.0

    """

    def __init__(self, window_size, ddof=1):
        self.ddof = ddof
        self.sos = 0
        self.rolling_mean = mean.RollingMean(window_size=window_size)

    @property
    def window_size(self):
        return self.rolling_mean.window_size

    def update(self, x):
        if len(self.rolling_mean) >= self.rolling_mean.size:
            self.sos -= self.rolling_mean[0] ** 2

        self.sos += x * x
        self.rolling_mean.update(x)
        return self

    @property
    def correction_factor(self):
        n = len(self.rolling_mean)
        if n > self.ddof:
            return n / (n - self.ddof)
        return 1

    def get(self):
        try:
            variance = (self.sos / len(self.rolling_mean)) - self.rolling_mean.get() ** 2
            return self.correction_factor * variance
        except ZeroDivisionError:
            return 0.



from .. import base


__all__ = ['StackingBinaryClassifier']


class StackingBinaryClassifier(base.Ensemble, base.BinaryClassifier):
    """Stacking for binary classification.

    Parameters:
        classifiers (list of `base.BinaryClassifier`)
        meta_classifier (`base.BinaryClassifier`)
        include_features (bool): Indicates whether or not the original features should be provided
            to the meta-model along with the predictions from each model.

    Example:

        ::

            >>> from creme import compose
            >>> from creme import ensemble
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_breast_cancer(),
            ...     shuffle=False
            ... )
            >>> model = compose.Pipeline([
            ...     ('scale', preprocessing.StandardScaler()),
            ...     ('stack', ensemble.StackingBinaryClassifier(
            ...         classifiers=[
            ...             linear_model.LogisticRegression(),
            ...             linear_model.PAClassifier(mode=1, C=0.01),
            ...             linear_model.PAClassifier(mode=2, C=0.01)
            ...         ],
            ...         meta_classifier=linear_model.LogisticRegression()
            ...     ))
            ... ])
            >>> metric = metrics.F1()

            >>> model_selection.online_score(X_y, model, metric)
            F1: 0.957123

    References:
        1. `A Kaggler's Guide to Model Stacking in Practice <http://blog.kaggle.com/2016/12/27/a-kagglers-guide-to-model-stacking-in-practice/>`_

    """

    def __init__(self, classifiers, meta_classifier, include_features=True):
        super().__init__(classifiers)
        self.meta_classifier = meta_classifier
        self.include_features = include_features

    def fit_one(self, x, y):

        # Ask each model to make a prediction and then update it
        oof = {}
        for i, classifier in enumerate(self):
            oof[f'oof_{i}'] = classifier.predict_proba_one(x).get(True, 0.5)
            classifier.fit_one(x, y)

        # Optionally, add the base features
        if self.include_features:
            oof.update(x)

        # Update the meta-classifier using the predictions from the base classifiers
        self.meta_classifier.fit_one(oof, y)

        return self

    def predict_proba_one(self, x):

        oof = {
            f'oof_{i}': classifier.predict_proba_one(x).get(True, 0.5)
            for i, classifier in enumerate(self)
        }

        if self.include_features:
            oof.update(x)

        return self.meta_classifier.predict_proba_one(oof)



import collections
import copy
import functools

from .. import base


class SelectKBest(base.Transformer):
    """Removes all but the $k$ highest scoring features.

    Parameters:
        similarity (stats.Bivariate)
        k (int)

    Attributes:
        similarities (dict): The similarity instances used for each feature.
        leaderboard (dict): The actual similarity measures.

    Example:

        ::

            >>> from pprint import pprint
            >>> from creme import feature_selection
            >>> from creme import stats
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X, y = datasets.make_regression(
            ...     n_samples=100,
            ...     n_features=10,
            ...     n_informative=2,
            ...     random_state=42
            ... )

            >>> selector = feature_selection.SelectKBest(
            ...     similarity=stats.PearsonCorrelation(),
            ...     k=2
            ... )

            >>> for xi, yi, in stream.iter_array(X, y):
            ...     selector = selector.fit_one(xi, yi)

            >>> pprint(selector.leaderboard)
            Counter({9: 0.789828...,
                     7: 0.544461...,
                     8: 0.106205...,
                     2: 0.063886...,
                     4: 0.053821...,
                     5: 0.027151...,
                     1: -0.031251...,
                     6: -0.065716...,
                     3: -0.150104...,
                     0: -0.189527...})

            >>> selector.transform_one(xi)
            {7: -1.279576..., 9: -1.840874...}

    """

    def __init__(self, similarity, k=10):
        self.k = k
        self.similarities = collections.defaultdict(functools.partial(copy.deepcopy, similarity))
        self.leaderboard = collections.Counter()

    @property
    def is_supervised(self):
        return True

    def fit_one(self, x, y):

        for i, xi in x.items():
            self.leaderboard[i] = self.similarities[i].update(xi, y).get()

        return self

    def transform_one(self, x):

        best_features = set(pair[0] for pair in self.leaderboard.most_common(self.k))

        if self.leaderboard:

            return {
                i: xi
                for i, xi in x.items()
                if i in best_features
            }

        return copy.deepcopy(x)



import collections
import copy

from sklearn import utils as sk_utils

from .. import base
from .. import optim
from .. import stats
from .. import utils


class FMRegressor(base.Regressor):
    """Factorization machine for regression.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import stream

            >>> features = [
            ... #     Users  |     Movies     |    Movie Ratings   | Time | Last Movies Rated
            ... #    A  B  C | TI  NH  SW  ST | TI   NH   SW   ST  |      | TI  NH  SW  ST
            ...     [1, 0, 0,  1,  0,  0,  0,   0.3, 0.3, 0.3, 0,     13,   0,  0,  0,  0 ],
            ...     [1, 0, 0,  0,  1,  0,  0,   0.3, 0.3, 0.3, 0,     14,   1,  0,  0,  0 ],
            ...     [1, 0, 0,  0,  0,  1,  0,   0.3, 0.3, 0.3, 0,     16,   0,  1,  0,  0 ],
            ...     [0, 1, 0,  0,  0,  1,  0,   0,   0,   0.5, 0.5,   5,    0,  0,  0,  0 ],
            ...     [0, 1, 0,  0,  0,  0,  1,   0,   0,   0.5, 0.5,   8,    0,  0,  1,  0 ],
            ...     [0, 0, 1,  1,  0,  0,  0,   0.5, 0,   0.5, 0,     9,    0,  0,  0,  0 ],
            ...     [0, 0, 1,  0,  0,  1,  0,   0.5, 0,   0.5, 0,     12,   1,  0,  0,  0 ]
            ... ]
            >>> target = [5, 3, 1, 4, 5, 1, 5]

            >>> fm = linear_model.FMRegressor(random_state=42)

            >>> for x, y in stream.iter_array(features[:5], target[:5]):
            ...     fm = fm.fit_one(x, y)

            >>> for x, y in stream.iter_array(features[5:], target[5:]):
            ...     print(fm.predict_one(x), y)
            -8.355825... 1
            -11.355220... 5

    """

    def __init__(self, n_components=5, optimizer=None, loss=None, intercept=None,
                 init_stdev=0.1, random_state=None):
        self.n_components = n_components
        self.loss = optim.losses.Squared() if loss is None else loss
        self.init_stdev = init_stdev
        self.random_state = sk_utils.check_random_state(random_state)

        self.weights = collections.defaultdict(float)
        self.latents = collections.defaultdict(self._make_random_latent_weights)
        if intercept is None or intercept is True:
            intercept = stats.Mean()
        self.intercept = intercept

        optimizer = optim.SGD(0.01) if optimizer is None else optimizer
        self.weights_optimizer = copy.deepcopy(optimizer)
        self.latents_optimizer = copy.deepcopy(optimizer)

    def _make_random_latent_weights(self):
        return {
            f: self.random_state.normal(scale=self.init_stdev)
            for f in range(self.n_components)
        }

    def fit_one(self, x, y):

        # Start by making a prediction
        y_pred = self.predict_one(x)

        # Compute the gradient of the loss with respect to the prediction
        loss_gradient = self.loss.gradient(y, y_pred)

        # Update the weights
        self.weights = self.weights_optimizer.update_after_pred(
            w=self.weights,
            g={j: xj * loss_gradient for j, xj in x.items()}
        )

        # Update the latent weights
        v = self.latents

        vs = {
            f: sum(
                v[j][f] * xj
                for j, xj in x.items()
            )
            for f in range(self.n_components)
        }

        for j, xj in x.items():
            self.latents[j] = self.latents_optimizer.update_after_pred(
                w=self.latents[j],
                g={
                    f: xj * (vs[f] - v[j][f] * xj)
                    for f in range(self.n_components)
                }
            )

        return self

    def predict_one(self, x):

        # For notational convenience
        w0 = self.intercept.get() if self.intercept else 0
        w = self.weights
        v = self.latents

        # Start with the intercept
        y_pred = w0

        # Add the unary interactions
        y_pred = utils.math.dot(x, w)

        # Add the pairwise interactions
        for f in range(self.n_components):
            y_pred += 0.5 * (
                sum(v[j][f] * xj for j, xj in x.items()) ** 2 -
                sum(v[j][f] ** 2 * xj ** 2 for j, xj in x.items())
            )

        return y_pred



import abc
import collections
import functools
import operator

from .. import base
from .. import proba
from .. import utils


def decimal_range(start, stop, num):
    """

    Example:

        >>> for x in decimal_range(1, 3, 5):
        ...     print(x)
        1
        1.5
        2.0
        2.5
        3.0

    """
    step = (stop - start) / (num - 1)

    while start <= stop:
        yield start
        start += step


class Op(collections.namedtuple('Op', 'symbol operator')):

    def __call__(self, a, b):
        return self.operator(a, b)

    def __str__(self):
        return self.symbol


LT = Op('<', operator.lt)
EQ = Op('=', operator.eq)


class Split(collections.namedtuple('Split', 'on how at')):
    """A data class for storing split details."""

    def __str__(self):
        at = self.at
        if isinstance(at, float):
            at = f'{at:.3f}'
        return f'{self.on} {self.how} {at}'

    def __call__(self, x):
        return self.how(x[self.on], self.at)


class SplitEnum(abc.ABC):

    def __init__(self, feature_name):
        self.feature_name = feature_name

    @abc.abstractmethod
    def update(self, x, y):
        """Updates the sufficient statistics used for evaluting splits."""

    @abc.abstractmethod
    def enumerate_splits(self):
        """Yields candidate split points and associated operators."""


class NumericSplitEnum(SplitEnum):
    """Split enumerator for classification and numerical attributes."""

    def __init__(self, feature_name, n_bins, n_splits):
        super().__init__(feature_name)
        self.P_xy = collections.defaultdict(functools.partial(utils.Histogram, max_bins=n_bins))
        self.n_splits = n_splits

    def update(self, x: float, y: 'base.Label'):
        self.P_xy[y].update(x)
        return self

    def enumerate_splits(self, target_dist: 'proba.Multinomial'):

        low = min(h[0].right for h in self.P_xy.values())
        high = min(h[-1].right for h in self.P_xy.values())

        # If only one single value has been observed, then no split can be proposed
        if low >= high:
            return
            yield

        for t in decimal_range(start=low, stop=high, num=self.n_splits):

            l_dist = {}
            r_dist = {}

            for y in target_dist:
                p_xy = self.P_xy[y].cdf(x=t) if y in self.P_xy else 0.  # P(x < t | y)
                p_y = target_dist.pmf(y)  # P(y)
                l_dist[y] = target_dist.n_samples * p_y * p_xy  # P(y | x < t)
                r_dist[y] = target_dist.n_samples * p_y * (1 - p_xy)  # P(y | x >= t)

            l_dist = proba.Multinomial(l_dist)
            r_dist = proba.Multinomial(r_dist)

            yield Split(on=self.feature_name, how=LT, at=t), l_dist, r_dist


class CategoricalSplitEnum(SplitEnum):
    """Split enumerator for classification and categorical attributes."""

    def __init__(self, feature_name):
        super().__init__(feature_name)
        self.P_xy = collections.defaultdict(proba.Multinomial)

    def update(self, x: str, y: 'base.Label'):
        self.P_xy[y].update(x)
        return self

    def enumerate_splits(self, target_dist: 'proba.Multinomial'):

        categories = set(*(p_x.keys() for p_x in self.P_xy.values()))

        # There has to be at least two categories for a split to be possible
        if len(categories) < 2:
            return
            yield

        for cat in categories:

            l_dist = {}
            r_dist = {}

            for y in target_dist:
                p_xy = self.P_xy[y].pmf(cat)  # P(cat | y)
                p_y = target_dist.pmf(y)  # P(y)
                l_dist[y] = target_dist.n_samples * p_y * p_xy  # P(y | cat)
                r_dist[y] = target_dist.n_samples * p_y * (1. - p_xy)  # P(y | !cat)

            l_dist = proba.Multinomial(l_dist)
            r_dist = proba.Multinomial(r_dist)

            yield Split(on=self.feature_name, how=EQ, at=cat), l_dist, r_dist



"""Base interfaces.

Every estimator in ``creme`` is a class, and as such inherits from at least one base interface.
These are used to categorize, organize, and standardize the many estimators that ``creme``
contains.

"""
import abc
import collections
import inspect
import typing

from . import utils


__all__ = [
    'BinaryClassifier',
    'Clusterer',
    'Ensemble',
    'Estimator',
    'Wrapper',
    'MultiClassifier',
    'MultiOutputClassifier',
    'MultiOutputRegressor',
    'Regressor',
    'Transformer',
    'OutlierDetector'
]


# Input and output types
Label = typing.Union[bool, str, int]
Proba = float
Probas = typing.Dict[Label, Proba]


DEFAULT_TAGS = {
    'handles_text': False,
    'requires_positive_data': False,
    'handles_categorical_features': False,
    'poor_score': False
}


def _update_if_consistent(dict1, dict2):
    common_keys = set(dict1.keys()).intersection(dict2.keys())
    for key in common_keys:
        if dict1[key] != dict2[key]:
            raise TypeError(f'Inconsistent values for tag {key}: {dict1[key]} != {dict2[key]}')
    dict1.update(dict2)
    return dict1


class Estimator:
    """An estimator."""

    def __str__(self):
        return self.__class__.__name__

    def __repr__(self):
        return utils.pretty_format_class(self)

    def _more_tags(self) -> dict:
        """Specific tags for this estimator."""
        return {}

    def _get_tags(self) -> dict:
        """Returns the estimator's tags."""

        tags = {}

        for base_class in inspect.getmro(self.__class__):
            if hasattr(base_class, '_more_tags') and base_class != self.__class__:
                tags = _update_if_consistent(tags, base_class._more_tags(self))

        if hasattr(self, '_more_tags'):
            tags = _update_if_consistent(tags, self._more_tags())

        return {**DEFAULT_TAGS, **tags}


class Regressor(Estimator):
    """A regressor."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: float) -> 'Regressor':
        """Fits to a set of features ``x`` and a real-valued target ``y``.

        Parameters:
            x (dict)
            y (float)

        Returns:
            self: object

        """

    @abc.abstractmethod
    def predict_one(self, x: dict) -> float:
        """Predicts the target value of a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            float: The prediction.

        """


class Classifier(Estimator):
    """A classifier."""

    @abc.abstractmethod
    def predict_proba_one(self, x: dict) -> Probas:
        """Predicts the probability output of a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            dict of floats

        """

    def predict_one(self, x: dict) -> Label:
        """Predicts the target value of a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            Label

        """
        y_pred = self.predict_proba_one(x)
        if y_pred:
            return max(y_pred, key=y_pred.get)
        return None


class BinaryClassifier(Classifier):
    """A binary classifier."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: bool) -> 'BinaryClassifier':
        """Fits to a set of features ``x`` and a boolean target ``y``.

        Parameters:
            x (dict)
            y (bool)

        Returns:
            self

        """


class MultiClassifier(BinaryClassifier):
    """A multi-class classifier."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y: Label) -> 'MultiClassifier':
        """Fits to a set of features ``x`` and a label ``y``.

        Parameters:
            x (dict)
            y (Label)

        Returns:
            self

        """


class Transformer(Estimator):
    """A transformer."""

    def fit_one(self, x: dict, y=None) -> 'Transformer':
        """Fits to a set of features ``x`` and an optional target ``y``.

        A lot of transformers don't actually have to do anything during the ``fit_one`` step
        because they are stateless. For this reason the default behavior of this function is to do
        nothing. Transformers that however do something during the ``fit_one`` can override this
        method.

        Parameters:
            x (dict)
            y (optional)

        Returns:
            self

        """
        return self

    @abc.abstractmethod
    def transform_one(self, x: dict) -> dict:
        """Transforms a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            dict

        """

    @property
    def is_supervised(self) -> bool:
        """Indicates if the transformer uses the target ``y`` or not.

        Supervised transformers have to be handled differently from unsupervised transformers in an
        online setting. This is especially true for target encoding where leakage can easily occur.
        Most transformers are unsupervised and so this property returns by default ``False``.
        Transformers that are supervised must override this property in their definition.

        """
        return False

    def __or__(self, other):
        """Merges with another Transformer into a Pipeline."""
        from . import compose
        if isinstance(other, compose.Pipeline):
            return other.__ror__(self)
        return compose.Pipeline([self, other])

    def __ror__(self, other):
        """Merges with another Transformer into a Pipeline."""
        from . import compose
        if isinstance(other, compose.Pipeline):
            return other.__or__(self)
        return compose.Pipeline([other, self])

    def __add__(self, other):
        """Merges with another Transformer into a TransformerUnion."""
        from . import compose
        if isinstance(other, compose.TransformerUnion):
            return other.__add__(self)
        return compose.TransformerUnion([self, other])

    def __radd__(self, other):
        """Merges with another Transformer into a TransformerUnion."""
        from . import compose
        if isinstance(other, compose.TransformerUnion):
            return other.__add__(self)
        return compose.TransformerUnion([other, self])


class Clusterer(Estimator):
    """A clusterer."""

    @abc.abstractmethod
    def fit_one(self, x: dict, y=None) -> 'Clusterer':
        """Fits to a set of features ``x``.

        Parameters:
            x (dict)
            y: Not used, only present for API consistency.

        Returns:
            self

        """

    @abc.abstractmethod
    def predict_one(self, x: dict) -> int:
        """Predicts the cluster number of a set of features ``x``.

        Parameters:
            x (dict)

        Returns:
            int

        """


class MultiOutputEstimator(Estimator):
    """A multi-output estimator."""


class MultiOutputClassifier(MultiOutputEstimator):
    """A multi-output classifier."""

    def fit_one(self, x: dict, y: typing.Dict[str, Label]) -> 'MultiOutputClassifier':
        """Fits to a set of features ``x`` and a set of labels ``y``.

        Parameters:
            x (dict)
            y (dict of Label)

        Returns:
            self

        """

    def predict_one(self, x: dict) -> typing.Dict[str, Label]:
        """Given a set of features ``x``, predicts a label for each output.

        Parameters:
            x (dict)

        Returns:

        """
        y_pred = self.predict_proba_one(x)
        return {
            c: max(y_pred[c], key=y_pred[c].get)
            for c in y_pred
        }

    def predict_proba_one(self, x: dict) -> typing.Dict[str, Probas]:
        """Given a set of features ``x``, predicts a the probability of each label for each output.

        Parameters:
            x (dict)

        Returns:

        """


class MultiOutputRegressor(MultiOutputEstimator):
    """A multi-output regressor."""

    def fit_one(self, x: dict, y: typing.Dict[str, float]) -> 'MultiOutputRegressor':
        """Fits to a set of features ``x`` and a set of outputs ``y``.

        Parameters:
            x (dict)
            y (dict of Label)

        Returns:
            self

        """

    def predict_one(self, x: dict) -> typing.Dict[str, float]:
        """Given a set of features ``x``, predicts a label for each output.

        Parameters:
            x (dict)

        Returns:
            dict

        """


class Wrapper(abc.ABC):
    """A wrapper model."""

    @abc.abstractproperty
    def _model(self):
        """Provides access to the wrapped model."""

    def __str__(self):
        return f'{type(self).__name__}({self._model})'


class Ensemble(Estimator, collections.UserList):
    """An ensemble model."""


class OutlierDetector(Estimator):

    @abc.abstractmethod
    def fit_one(self, x: dict) -> 'OutlierDetector':
        """Updates the model."""

    @abc.abstractmethod
    def score_one(self, x: dict) -> float:
        """Returns an outlier score.

        The range of the score depends on each model. Some models will output anomaly scores
        between 0 and 1, others will not. In any case, the lower the score, the more likely it is
        that ``x`` is an anomaly.

        """



import math

from .. import base


__all__ = ['TargetModifierRegressor', 'BoxCoxTransformRegressor']


class TargetModifierRegressor(base.Regressor, base.Wrapper):
    """Model wrapper that modifies the target before training.

    The user is expected to check that ``func`` and ``inverse_func`` are coherent with each other.

    Parameters:
        regressor (creme.base.Regressor): regressor model applied for training.
        func (callable): a function modifying the target before training.
        inverse_func (callable): a function to return to the target's original space.

    Example:

        ::

            >>> import math
            >>> from creme import compose
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_boston(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> model = compose.Pipeline([
            ...     ('scale', preprocessing.StandardScaler()),
            ...     ('learn', compose.TargetModifierRegressor(
            ...         regressor=linear_model.LinearRegression(intercept_lr=0.15),
            ...         func=math.log,
            ...         inverse_func=math.exp
            ...     ))
            ... ])
            >>> metric = metrics.MSE()

            >>> model_selection.online_score(X_y, model, metric)
            MSE: 37.499423

    """

    def __init__(self, regressor, func, inverse_func):
        self.regressor = regressor
        self.func = func
        self.inverse_func = inverse_func

    @property
    def _model(self):
        return self.regressor

    def fit_one(self, x, y):
        self.regressor.fit_one(x, self.func(y))
        return self

    def predict_one(self, x):
        y_pred = self.regressor.predict_one(x)
        return self.inverse_func(y_pred)


class BoxCoxTransformRegressor(TargetModifierRegressor):
    """Applies the Box-Cox transform to the target before training.

    Box-Cox transform is useful when the target variable is heteroscedastic (i.e. there are
    sub-populations that have different variabilities from others) allowing to transform it towards
    normality.

    The ``power`` parameter is denoted λ in the litterature. If ``power`` is equal to 0 than the
    Box-Cox transform will be equivalent to a log transform.

    Parameter:
        regressor (creme.base.Regressor): regressor model applied for training.
        power (float): power value to do the transformation.

    Example:

        ::

            >>> import math
            >>> from creme import compose
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_boston(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> model = compose.Pipeline([
            ...     ('scale', preprocessing.StandardScaler()),
            ...     ('learn', compose.BoxCoxTransformRegressor(
            ...         regressor=linear_model.LinearRegression(intercept_lr=0.2),
            ...         power=0.05
            ...     ))
            ... ])
            >>> metric = metrics.MSE()

            >>> model_selection.online_score(X_y, model, metric)
            MSE: 37.225174

    """

    def __init__(self, regressor, power=1.):
        super().__init__(
            regressor=regressor,
            func=(lambda y: (y ** power - 1) / power) if power > 0 else math.log,
            inverse_func=(lambda y: (power * y + 1) ** (1 / power)) if power > 0 else math.exp
        )



from .. import base


__all__ = ['Blacklister']


class Blacklister(base.Transformer):
    """Subsets a set of features by applying a blacklist.

    Parameters:
        blacklist (strs): Key(s) to discard.

    Example:

        ::

            >>> from creme import compose

            >>> x = {'a': 42, 'b': 12}
            >>> compose.Blacklister('a', 'zoidberg').transform_one(x)
            {'b': 12}

            >>> compose.Blacklister('b').transform_one(x)
            {'a': 42}

    """

    def __init__(self, *blacklist):
        self.blacklist = set(blacklist)

    def transform_one(self, x):
        return {i: x[i] for i in set(x.keys()) - self.blacklist}

    def __str__(self):
        return '~' + str(sorted(self.blacklist))



from . import base


class EWMean(base.Univariate):
    """Exponentially weighted mean.

    Parameters:
        alpha (float): The closer ``alpha`` is to 1 the more the statistic will adapt to recent
            values.

    Attributes:
        mean (float) : The running exponentially weighted mean.

    Example:

        ::

            >>> from creme import stats

            >>> X = [1, 3, 5, 4, 6, 8, 7, 9, 11]
            >>> ewm = stats.EWMean(alpha=0.5)
            >>> for x in X:
            ...     print(ewm.update(x).get())
            1
            2.0
            3.5
            3.75
            4.875
            6.4375
            6.71875
            7.859375
            9.4296875

    References:
        1. `Incremental calculation of weighted mean and variance <http://people.ds.cam.ac.uk/fanf2/hermes/doc/antiforgery/stats.pdf>`_
        2. `Exponential Moving Average on Streaming Data <https://dev.to/nestedsoftware/exponential-moving-average-on-streaming-data-4hhl>`_

    """

    def __init__(self, alpha=0.5):
        self.alpha = alpha
        self.mean = 0

    @property
    def name(self):
        return f'ewm_{self.alpha}'

    def update(self, x):
        self.mean = self.alpha * x + (1. - self.alpha) * self.mean if self.mean else x
        return self

    def get(self):
        return self.mean



import collections
import copy
import math

from sklearn import utils

from .. import base


__all__ = ['AdaBoostClassifier']


class BaseBoosting(base.Wrapper, base.Ensemble):

    def __init__(self, model, n_models=10, random_state=None):
        super().__init__(copy.deepcopy(model) for i in range(n_models))
        self.model = model
        self.rng = utils.check_random_state(random_state)

    @property
    def _model(self):
        return self.model


class AdaBoostClassifier(base.Classifier, BaseBoosting):
    '''Boosting for classification

    For each incoming observation, each model's ``fit_one`` method is called ``k`` times where
    ``k`` is sampled from a Poisson distribution of parameter lambda. The lambda parameter is 
    updated when the weaks learners fit successively the same observation. 

    Parameters:
        model (BinaryClassifier or MultiClassifier): The classifier to boost.
        n_models (int): The number of models in the ensemble.
        random_state (int, ``numpy.random.RandomState`` instance or None): If int, ``random_state``
            is the seed used by the random number generator; if ``RandomState`` instance,
            ``random_state`` is the random number generator; if ``None``, the random number
            generator is the ``RandomState`` instance used by `numpy.random`.

    Attributes:
        wrong_weight (collections.defaultdict): Number of times a model has made a mistake
            when making predictions.
        correct_weight (collections.defaultdict): Number of times a model has predicted the right
            label when making predictions.


    Example:

        In the following example three tree classifiers are boosted together. The performance is
        slightly better than when using a single tree.

        ::

            >>> from creme import datasets
            >>> from creme import ensemble
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import tree

            >>> X_y = datasets.fetch_electricity()

            >>> metric = metrics.LogLoss()

            >>> model = ensemble.AdaBoostClassifier(
            ...     model=(
            ...         tree.DecisionTreeClassifier(
            ...             criterion='gini', 
            ...             confidence=1e-5, 
            ...             patience=2000
            ...         )
            ...     ),
            ...     n_models=5,
            ...     random_state=42
            ... )

            >>> model_selection.online_score(X_y, model, metric)
            LogLoss: 0.5531

            >>> print(model)
            AdaBoostClassifier(DecisionTreeClassifier)

    References:
        1. `Online Bagging and Boosting <https://ti.arc.nasa.gov/m/profile/oza/files/ozru01a.pdf>`_
        2. `Github repository for online boosting <https://github.com/crm416/online_boosting/blob/master/ensemblers/adaboost.py>`_

    '''

    def __init__(self, model, n_models=10, random_state=None):
        super().__init__(model, n_models, random_state)
        self.wrong_weight = collections.defaultdict(int)
        self.correct_weight = collections.defaultdict(int)

    def fit_one(self, x, y):  
        lambda_poisson = 1
        
        for i, model in enumerate(self):
            for _ in range(self.rng.poisson(lambda_poisson)):
                model.fit_one(x, y)
            
            if model.predict_one(x) == y:
                self.correct_weight[i] += lambda_poisson
                lambda_poisson *= (
                    (self.correct_weight[i] + self.wrong_weight[i]) / (2 * self.correct_weight[i]) 
                )
            
            else:
                self.wrong_weight[i] += lambda_poisson
                lambda_poisson *= (
                    (self.correct_weight[i] + self.wrong_weight[i]) / (2 * self.wrong_weight[i])
                )
        return self


    def predict_proba_one(self, x):
        """
        Store the predicted probabilities with the corresponding weights for each weak learner and 
        return the probabilities associated to the model which has maximum weight.
        """
        model_weights = collections.defaultdict(int)
        predictions   = {}

        for i, model in enumerate(self):
            epsilon = self.correct_weight[i] + 1e-16
            epsilon /= (self.wrong_weight[i]  + 1e-16)
            weight = math.log(epsilon)
            model_weights[i] += weight 
            predictions[i] = model.predict_proba_one(x)

        y_pred = predictions[max(model_weights, key=model_weights.get)]
        total = sum(y_pred.values())
        return {label: proba / total for label, proba in y_pred.items()}



import collections
import math

from .. import utils

from . import base


__all__ = ['AdaBound']


class AdaBound(base.Optimizer):
    """AdaBound optimizer.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_breast_cancer(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> optimizer = optim.AdaBound()
            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LogisticRegression(optimizer)
            ... )
            >>> metric = metrics.F1()

            >>> model_selection.online_score(X_y, model, metric)
            F1: 0.963889

    References:
        1. `Adaptive Gradient Methods with Dynamic Bound of Learning Rate <https://openreview.net/forum?id=Bkg3g2R9FX>`_

    """

    def __init__(self, lr=1e-3, beta_1=0.9, beta_2=0.999, eps=1e-8, gamma=1e-3, final_lr=0.1):
        super().__init__(lr)
        self.base_lr = lr
        self.final_lr = final_lr
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.eps = eps
        self.gamma = gamma
        self.m = collections.defaultdict(float)
        self.v = collections.defaultdict(float)

    def _update_after_pred(self, w, g):

        for i, gi in g.items():
            self.m[i] = self.beta_1 * self.m[i] + (1 - self.beta_1) * gi
            self.v[i] = self.beta_2 * self.v[i] + (1 - self.beta_2) * gi ** 2

            bias_1 = 1 - self.beta_1 ** (self.n_iterations + 1)
            bias_2 = 1 - self.beta_2 ** (self.n_iterations + 1)
            step_size = self.learning_rate * math.sqrt(bias_2) / bias_1
            step_size_bound = step_size / (math.sqrt(self.v[i]) + self.eps)

            self.final_lr *= self.learning_rate / self.base_lr
            lower_bound = self.final_lr * (1 - 1 / (self.gamma * (self.n_iterations + 1) + 1))
            upper_bound = self.final_lr * (1 + 1 / (self.gamma * (self.n_iterations + 1)))

            w[i] -= utils.math.clamp(step_size_bound, lower_bound, upper_bound) * self.m[i]

        return w



from . import mse


__all__ = ['RMSE']


class RMSE(mse.MSE):
    """Root mean squared error.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [3, -0.5, 2, 7]
            >>> y_pred = [2.5, 0.0, 2, 8]

            >>> metric = metrics.RMSE()
            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(metric.update(y_t, y_p).get())
            0.5
            0.5
            0.408248...
            0.612372...

            >>> metric
            RMSE: 0.612372

    """

    def get(self):
        return super().get() ** .5



import collections
import math
import re

from sklearn import feature_extraction

from .. import base


__all__ = ['CountVectorizer', 'TFIDFVectorizer']


class VectorizerMixin:
    """Contains common processing steps used by each vectorizer.

    Parameters:
        on (str): The name of the feature that contains the text to vectorize. If `None`, then each
            ``fit_one`` and ``transform_one`` should treat ``x`` as a `str` and not as a ``dict``.
        strip_accents (bool): Whether or not to strip accent characters.
        lowercase (bool): Whether or not to convert all characters to lowercase.
        tokenizer (callable): The function used to convert preprocessed text into a `dict` of
            tokens. A default one is used if it is not provided by the user.

    """

    def __init__(self, on=None, strip_accents=True, lowercase=True, tokenizer=None):
        self.on = on
        self.strip_accents = strip_accents
        self.lowercase = lowercase
        self.tokenizer = re.compile(r'(?u)\b\w\w+\b').findall if tokenizer is None else tokenizer

    def _get_text(self, x):
        if self.on is not None:
            return x[self.on]
        return x

    def _more_tags(self):
        return {'handles_text': True}

    def preprocess(self, text):
        """Returns a function to preprocess the text before tokenization."""

        if self.strip_accents:
            text = feature_extraction.text.strip_accents_unicode(text)

        if self.lowercase:
            text = text.lower()

        return text


class CountVectorizer(base.Transformer, VectorizerMixin):
    """Counts the number of occurrences of each token.

    This returns exactly the same results as `sklearn.feature_extraction.text.CountVectorizer`.

    Parameters:
        on (str): The name of the feature that contains the text to vectorize. If ``None``, then
            the input is treated as a document instead of a set of features.
        strip_accents (bool): Whether or not to strip accent characters.
        lowercase (bool): Whether or not to convert all characters to lowercase.
        tokenizer (callable): The function used to convert preprocessed text into a `dict` of
            tokens. A default one is used if it is not provided by the user.

    Example:

        ::

            >>> import creme
            >>> corpus = [
            ...     'This is the first document.',
            ...     'This document is the second document.',
            ...     'And this is the third one.',
            ...     'Is this the first document?',
            ... ]
            >>> vectorizer = creme.feature_extraction.CountVectorizer()
            >>> for sentence in corpus:
            ...     print(vectorizer.transform_one(sentence))
            Counter({'this': 1, 'is': 1, 'the': 1, 'first': 1, 'document': 1})
            Counter({'document': 2, 'this': 1, 'is': 1, 'the': 1, 'second': 1})
            Counter({'and': 1, 'this': 1, 'is': 1, 'the': 1, 'third': 1, 'one': 1})
            Counter({'is': 1, 'this': 1, 'the': 1, 'first': 1, 'document': 1})

    """

    def transform_one(self, x):
        return collections.Counter(self.tokenizer(self.preprocess(self._get_text(x))))


class TFIDFVectorizer(base.Transformer, VectorizerMixin):
    """Computes values TF-IDF values.

    We use the same definition as scikit-learn. The only difference in the results comes the fact
    that the document frequencies have to be computed online.

    Parameters:
        on (str): The name of the feature that contains the text to vectorize. If ``None``, then
            the input is treated as a document instead of a set of features.
        strip_accents (bool): Whether or not to strip accent characters.
        lowercase (bool): Whether or not to convert all characters to lowercase.
        tokenizer (callable): The function used to convert preprocessed text into a `dict` of
            tokens. A default one is used if it is not provided by the user.
        normalize (bool): Whether or not the TF-IDF values by their L2 norm.

    Attributes:
        tfs (feature_extraction.CountVectorizer): The term counts.
        dfs (collections.defaultdict): The document counts.
        n (int): The number of scanned documents.

    Example:

        ::

            >>> import creme
            >>> corpus = [
            ...     'This is the first document.',
            ...     'This document is the second document.',
            ...     'And this is the third one.',
            ...     'Is this the first document?',
            ... ]
            >>> vectorizer = creme.feature_extraction.TFIDFVectorizer()
            >>> for sentence in corpus:
            ...     print(vectorizer.fit_one(sentence).transform_one(sentence))
            {'this': 0.447..., 'is': 0.447..., 'the': 0.447..., 'first': 0.447..., 'document': 0.447...}
            {'this': 0.333..., 'document': 0.667..., 'is': 0.333..., 'the': 0.333..., 'second': 0.469...}
            {'and': 0.497..., 'this': 0.293..., 'is': 0.293..., 'the': 0.293..., 'third': 0.497..., 'one': 0.497...}
            {'is': 0.384..., 'this': 0.384..., 'the': 0.384..., 'first': 0.580..., 'document': 0.469...}

    """

    def __init__(self, on=None, strip_accents=True, lowercase=True, tokenizer=None,
                 normalize=True):
        super().__init__(on, strip_accents, lowercase, tokenizer)
        self.normalize = normalize
        self.tfs = CountVectorizer(on, strip_accents, lowercase, tokenizer)
        self.dfs = collections.defaultdict(int)
        self.n = 0

    def compute_tfidfs(self, term_counts):
        n_terms = sum(term_counts.values())

        tfidfs = {}

        for term, count in term_counts.items():
            tf = count / n_terms
            idf = math.log(((1 + self.n) / (1 + self.dfs[term]))) + 1
            tfidfs[term] = tf * idf

        if self.normalize:
            norm = math.sqrt(sum(tfidf ** 2 for tfidf in tfidfs.values()))
            return {term: tfidf / norm for term, tfidf in tfidfs.items()}
        return tfidfs

    def fit_one(self, x, y=None):

        text = self._get_text(x)

        # Compute the term counts
        term_counts = self.tfs.fit_one(text).transform_one(text)

        # Increment the document frequencies of each term
        for term in term_counts:
            self.dfs[term] += 1

        # Increment the global document counter
        self.n += 1

        return self

    def transform_one(self, x):
        term_counts = self.tfs.transform_one(self._get_text(x))
        return self.compute_tfidfs(term_counts)



import collections

from . import base


__all__ = ['AdaDelta']


class AdaDelta(base.Optimizer):
    """AdaDelta optimizer.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_breast_cancer(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> optimizer = optim.AdaDelta()
            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LogisticRegression(optimizer)
            ... )
            >>> metric = metrics.F1()

            >>> model_selection.online_score(X_y, model, metric)
            F1: 0.946092

    References:

        1. `AdaDelta: an adaptive learning rate method <https://arxiv.org/pdf/1212.5701.pdf>`_

    """

    def __init__(self, rho=0.95, eps=1e-8):
        super().__init__(lr=None)
        self.rho = rho
        self.eps = eps
        self.g2 = collections.defaultdict(float)
        self.s2 = collections.defaultdict(float)

    def _rms(self, x):
        return (x + self.eps) ** 0.5

    def _update_after_pred(self, w, g):

        for i, gi in g.items():

            # Accumulate the gradient
            self.g2[i] = self.rho * self.g2[i] + (1 - self.rho) * gi ** 2

            # Compute the update
            step = - self._rms(self.s2[i]) / self._rms(self.g2[i]) * gi

            # Accumulate the update
            self.s2[i] = self.rho * self.s2[i] + (1 - self.rho) * step ** 2

            # Apply the update
            w[i] += step

        return w



import collections
import numbers

from .. import base
from .. import optim
from .. import utils


class GLM:
    """Generalized Linear Model.

    Parameters:
        optimizer (optim.Optimizer): The sequential optimizer used for updating the weights. Note
            that the intercept is handled separately.
        loss (optim.Loss): The loss function to optimize for.
        intercept (float): Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated. Setting this to 0 means that no intercept will be used.
        l2 (float): Amount of L2 regularization used to push weights towards 0.
        clip_gradient (float): Clips the absolute value of each gradient value.

    Attributes:
        weights (collections.defaultdict)

    """

    def __init__(self, optimizer, loss, l2, intercept, intercept_lr, clip_gradient):
        self.optimizer = optimizer
        self.loss = loss
        self.l2 = l2
        self.intercept = intercept
        self.intercept_lr = (
            optim.schedulers.Constant(intercept_lr)
            if isinstance(intercept_lr, numbers.Number) else
            intercept_lr
        )
        self.clip_gradient = clip_gradient
        self.weights = collections.defaultdict(float)

    def _raw_dot(self, x):
        return utils.math.dot(self.weights, x) + self.intercept

    def fit_one(self, x, y, sample_weight=1.):

        # Some optimizers need to do something before a prediction is made
        self.weights = self.optimizer.update_before_pred(w=self.weights)

        # Obtain the gradient of the loss with respect to the raw output
        g_loss = self.loss.gradient(y_true=y, y_pred=self._raw_dot(x))

        # Clamp the gradient to avoid numerical instability
        g_loss = utils.math.clamp(g_loss, minimum=-self.clip_gradient, maximum=self.clip_gradient)

        # Apply the sample weight
        g_loss *= sample_weight

        # Calculate the gradient
        gradient = {
            i: xi * g_loss + 2. * self.l2 * self.weights.get(i, 0)
            for i, xi in x.items()
        }

        # Update the intercept
        self.intercept -= self.intercept_lr.get(self.optimizer.n_iterations) * g_loss

        # Update the weights
        self.weights = self.optimizer.update_after_pred(w=self.weights, g=gradient)

        return self


class LinearRegression(GLM, base.Regressor):
    """Linear regression.

    Parameters:
        optimizer (optim.Optimizer): The sequential optimizer used for updating the weights. Note
            that the intercept is handled separately. Defaults to ``optim.SGD(.01)``.
        loss (optim.RegressionLoss): The loss function to optimize for. Defaults to
            ``optim.SquaredLoss``.
        intercept (float): Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated.
        l2 (float): Amount of L2 regularization used to push weights towards 0.

    Attributes:
        weights (collections.defaultdict): The current weights assigned to the features.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_boston(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LinearRegression(intercept_lr=.1)
            ... )
            >>> metric = metrics.MAE()

            >>> model_selection.online_score(X_y, model, metric)
            MAE: 4.038404

            >>> model['LinearRegression'].intercept
            22.189736...

    Note:
        Using a feature scaler such as `preprocessing.StandardScaler` upstream helps the optimizer
        to converge.

    """

    def __init__(self, optimizer=None, loss=None, l2=.0001, intercept=0., intercept_lr=.01,
                 clip_gradient=1e12):
        super().__init__(
            optimizer=(
                optim.SGD(optim.schedulers.InverseScaling(.01, .25))
                if optimizer is None else
                optimizer
            ),
            loss=optim.losses.Squared() if loss is None else loss,
            intercept=intercept,
            intercept_lr=intercept_lr,
            l2=l2,
            clip_gradient=clip_gradient
        )

    def predict_one(self, x):
        return self._raw_dot(x)


class LogisticRegression(GLM, base.BinaryClassifier):
    """Logistic regression.

    Parameters:
        optimizer (optim.Optimizer): The sequential optimizer used for updating the weights. Note
            that the intercept is handled separately. Defaults to ``optim.SGD(.05)``.
        loss (optim.BinaryLoss): The loss function to optimize for. Defaults to ``optim.LogLoss``.
        intercept (float): Initial intercept value.
        intercept_lr (optim.schedulers.Scheduler or float): Learning rate scheduler used for
            updating the intercept. If a `float` is passed, then an instance of
            `optim.schedulers.Constant` will be used. Setting this to 0 implies that the intercept
            will be not be updated.
        l2 (float): Amount of L2 regularization used to push weights towards .

    Attributes:
        weights (collections.defaultdict)

    Example:

        ::

            >>> from creme import datasets
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing

            >>> X_y = datasets.fetch_electricity()

            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LogisticRegression(optimizer=optim.SGD(.1))
            ... )
            >>> metric = metrics.Accuracy()

            >>> model_selection.online_score(X_y, model, metric)
            Accuracy: 0.894642

    Note:
        Using a feature scaler such as `preprocessing.StandardScaler` upstream helps the optimizer
        to converge.

    """

    def __init__(self, optimizer=None, loss=None, l2=.0001, intercept=0., intercept_lr=.01,
                 clip_gradient=1e12):
        super().__init__(
            optimizer=optim.SGD(.01) if optimizer is None else optimizer,
            loss=optim.losses.Log() if loss is None else loss,
            intercept=intercept,
            intercept_lr=intercept_lr,
            l2=l2,
            clip_gradient=clip_gradient
        )

    def predict_proba_one(self, x):
        p = utils.math.sigmoid(self._raw_dot(x))  # Convert log-odds ratio to probability
        return {False: 1. - p, True: p}



import collections

from sklearn import utils

from .. import base


__all__ = ['FeatureHasher']


class FeatureHasher(base.Transformer):
    """Implements the hashing trick.

    Each pair of (name, value) features is hashed into a random integer. A module operator is then
    used to make sure the hash is in a certain range. We use the Murmurhash implementation from
    scikit-learn.

    Parameters:
        n_features (int): The number by which each hash will be moduloed by.
        alternate_sign (bool): Whether or not half of the hashes will be negated.
        encoding (str): The string encoding used by `sklearn.utils.murmurhash3_32`.
        random_state (int, ``numpy.random.RandomState`` instance or None): If int, ``random_state``
            is the seed used by the random number generator; if ``RandomState`` instance,
            ``random_state`` is the random number generator; if ``None``, the random number
            generator is the ``RandomState`` instance used by `numpy.random`.

    Arguments:
        seed (int): Seed used by `sklearn.utils.murmurhash3_32`.

    Example:

        ::

            >>> import creme
            >>> hasher = creme.preprocessing.FeatureHasher(n_features=10, random_state=0)
            >>> X = [{'dog': 1, 'cat': 2, 'elephant': 4}, {'dog': 2, 'run': 5}]
            >>> for x in X:
            ...     print(hasher.transform_one(x))
            defaultdict(<class 'int'>, {2: -1, 1: 2, 8: 4})
            defaultdict(<class 'int'>, {2: -2, 4: 5})

    References:

        1. `Feature vectorization using hashing trick <https://www.wikiwand.com/en/Feature_hashing#/Feature_vectorization_using_hashing_trick>`_

    """

    def __init__(self, n_features=1048576, alternate_sign=True, encoding='utf-8',
                 random_state=None):
        self.n_features = n_features
        self.alternate_sign = alternate_sign
        self.encoding = encoding
        self.random_state = utils.check_random_state(random_state)
        self.seed = self.random_state.randint(0, 2 ** 32 - 1)

    def _hash_str(self, s):
        return utils.murmurhash3_32(bytes(s, self.encoding), self.seed)

    def transform_one(self, x):

        x_hashed = collections.defaultdict(int)

        for feature, value in x.items():

            if isinstance(value, str):
                feature = f'{feature}={value}'
                value = 1

            if value == 0:
                continue

            h = self._hash_str(feature)
            i = h % self.n_features

            if self.alternate_sign:
                value *= (h >= 0) * 2 - 1

            x_hashed[i] += value

        return x_hashed



from .. import stats

from . import base


__all__ = ['Accuracy']


class BaseAccuracy(base.MultiClassMetric):

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True


class Accuracy(stats.Mean, BaseAccuracy):
    """Accuracy score, which is the percentage of exact matches.

    Example:

        ::

            >>> import math
            >>> from creme import metrics
            >>> from sklearn.metrics import accuracy_score

            >>> y_true = [True, False, True, True, True]
            >>> y_pred = [True, True, False, True, True]

            >>> metric = metrics.Accuracy()
            >>> for i, (y_t, y_p) in enumerate(zip(y_true, y_pred)):
            ...     metric = metric.update(y_t, y_p)
            ...     assert math.isclose(metric.get(), accuracy_score(y_true[:i+1], y_pred[:i+1]))

            >>> metric
            Accuracy: 0.6

    """

    def update(self, y_true, y_pred, sample_weight=1.):
        return super().update(x=y_true == y_pred, w=sample_weight)

    def revert(self, y_true, y_pred, sample_weight=1.):
        return super().revert(x=y_true == y_pred, w=sample_weight)



import abc
import numbers
import typing

from . import schedulers


class Optimizer(abc.ABC):

    def __init__(self, lr: typing.Union[schedulers.Scheduler, float]):
        self.lr = (
            schedulers.Constant(lr)
            if isinstance(lr, numbers.Number) else
            lr
        )
        self.n_iterations = 0

    @property
    def learning_rate(self) -> float:
        return self.lr.get(self.n_iterations)

    @classmethod
    def update_before_pred(self, w: dict) -> dict:
        """Updates a weight vector before a prediction is made."""
        return w

    def _update_after_pred(self, w: dict, g: dict) -> dict:
        """Updates a weight vector given a gradient."""
        raise NotImplementedError

    def update_after_pred(self, w: dict, g: dict) -> dict:

        # Update the weights
        w = self._update_after_pred(w, g)

        # Update the iteration counter
        self.n_iterations += 1

        return w

    def __str__(self):
        return self.__class__.__name__



import collections
import itertools
import types

try:
    import graphviz
    GRAPHVIZ_INSTALLED = True
except ImportError:
    GRAPHVIZ_INSTALLED = False

from sklearn.utils import metaestimators

from .. import base

from . import func
from . import union


__all__ = ['Pipeline']


class Network(collections.UserList):
    """An abstraction to help with drawing pipelines."""

    def __init__(self, nodes, links, directed, name=None):
        super().__init__()
        for node in nodes:
            self.append(node)
        self.links = set()
        for link in links:
            self.link(*link)
        self.directed = directed
        self.name = name

    def append(self, a):
        if a not in self:
            super().append(a)

    def link(self, a, b):
        self.append(a)
        self.append(b)
        self.links.add((self.index(a), self.index(b)))

    def draw(self):
        dot = graphviz.Digraph()

        def draw_node(a):
            if isinstance(a, Network):
                for part in a:
                    draw_node(part)
            else:
                dot.node(a)

        for a in self:
            draw_node(a)

        def draw_link(a, b):

            if isinstance(a, Network):
                # Connect the last part of a with b
                if a.directed:
                    draw_link(a[-1], b)
                # Connect each part of a with b
                else:
                    for part in a:
                        draw_link(part, b)

            elif isinstance(b, Network):
                # Connect the first part of b with a
                if b.directed:

                    if b.name is not None:
                        # If the graph has a name, then we treat is as a cluster
                        c = b.draw()
                        c.attr(label=b.name)
                        c.name = f'cluster_{b.name}'
                        dot.subgraph(c)
                    else:
                        dot.subgraph(b.draw())

                    draw_link(a, b[0])
                # Connect each part of b with a
                else:
                    for part in b:
                        draw_link(a, part)

            else:
                dot.edge(a, b)

        for a, b in self.links:
            draw_link(self[a], self[b])

        return dot


class Pipeline(collections.OrderedDict):
    """Chains a sequence of estimators.

    Sequentially apply a list of estimators. Pipelines helps to define machine learning systems in a
    declarative style, which makes a lot of sense when we think in a stream manner. For further
    information and practical examples, take a look at the
    `user guide <../notebooks/the-art-of-using-pipelines.html>`_.

    Parameters:
        steps (list): Ideally a list of (name, estimator) tuples. If an estimator is given without
            a name then a name is automatically inferred from the estimator.

    Example:

        ::

            >>> from creme import compose
            >>> from creme import feature_extraction
            >>> from creme import linear_model
            >>> from creme import preprocessing

            >>> tfidf = feature_extraction.TFIDFVectorizer('text')
            >>> counts = feature_extraction.CountVectorizer('text')
            >>> text_part = compose.Whitelister('text') | (tfidf + counts)

            >>> num_part = compose.Whitelister('a', 'b') | preprocessing.PolynomialExtender()

            >>> model = text_part + num_part
            >>> model |= preprocessing.StandardScaler()
            >>> model |= linear_model.LinearRegression()

            >>> dot = model.draw()

        .. image:: ../_static/pipeline_docstring.svg
            :align: center

        The following shows an example of using ``debug_one`` to visualize how the information
        flows and changes throughout the pipeline.

        ::

            >>> from creme import compose
            >>> from creme import feature_extraction
            >>> from creme import naive_bayes
            >>> from sklearn import datasets

            >>> X_y = [
            ...     ('A positive comment', True),
            ...     ('A negative comment', False),
            ...     ('A happy comment', True),
            ...     ('A lovely comment', True),
            ...     ('A harsh comment', False)
            ... ]

            >>> tfidf = feature_extraction.TFIDFVectorizer() | compose.Renamer(prefix='tfidf_')
            >>> counts = feature_extraction.CountVectorizer() | compose.Renamer(prefix='count_')
            >>> mnb = naive_bayes.MultinomialNB()
            >>> model = (tfidf + counts) | mnb

            >>> for x, y in X_y:
            ...     model = model.fit_one(x, y)

            >>> model.debug_one(X_y[0][0])
            0. Input
            --------
            A positive comment
            <BLANKLINE>
            1. Transformer union
            --------------------
                1.0 TFIDFVectorizer | Renamer
                -----------------------------
                tfidf_comment: 0.47606 (float)
                tfidf_positive: 0.87942 (float)
            <BLANKLINE>
                1.1 CountVectorizer | Renamer
                -----------------------------
                count_comment: 1 (int)
                count_positive: 1 (int)
            <BLANKLINE>
            count_comment: 1 (int)
            count_positive: 1 (int)
            tfidf_comment: 0.50854 (float)
            tfidf_positive: 0.86104 (float)
            <BLANKLINE>
            2. MultinomialNB
            ----------------
            False: 0.19313
            True: 0.80687

    """

    def __init__(self, steps=None):
        if steps is not None:
            for step in steps:
                self |= step

    def __or__(self, other):
        """Inserts a step at the end of the pipeline."""
        self.add_step(other, at_start=False)
        return self

    def __ror__(self, other):
        """Inserts a step at the start of the pipeline."""
        self.add_step(other, at_start=True)
        return self

    def __add__(self, other):
        """Merges with another Pipeline or TransformerUnion into a TransformerUnion."""
        if isinstance(other, union.TransformerUnion):
            return other.__add__(self)
        return union.TransformerUnion([self, other])

    def __str__(self):
        """Return a human friendly representation of the pipeline."""
        return ' | '.join(self.keys())

    def __repr__(self):
        return (
            'Pipeline (\n    ' +
            '    '.join(',\n'.join(map(repr, self.values())).splitlines(True)) +
            '\n)'
        )

    @property
    def __class__(self):
        return self.final_estimator.__class__

    @property
    def transformers(self):
        """If a pipeline has $n$ steps, then the first $n-1$ are necessarily transformers."""
        if isinstance(self.final_estimator, base.Transformer):
            return self.values()
        return itertools.islice(self.values(), len(self) - 1)

    @property
    def is_supervised(self):
        """Only works if all the steps of the pipelines are transformers."""
        return any(transformer.is_supervised for transformer in self.values())

    def add_step(self, step, at_start):
        """Adds a step to either end of the pipeline while taking care of the input type."""

        # Infer a name if none is given
        if not isinstance(step, (list, tuple)):
            step = (str(step), step)

        name, estimator = step

        # If a function is given then wrap it in a FuncTransformer
        if isinstance(estimator, types.FunctionType):
            name = estimator.__name__
            estimator = func.FuncTransformer(estimator)

        # Check if an identical step has already been inserted
        if name in self:
            raise KeyError(f'{name} already exists')

        # Instantiate the estimator if it hasn't been done
        if isinstance(estimator, type):
            estimator = estimator()

        # Store the step
        self[name] = estimator

        # Move the step to the start of the pipeline if so instructed
        if at_start:
            self.move_to_end(name, last=False)

    @property
    def final_estimator(self):
        """The final estimator."""
        return self[next(reversed(self))]

    def fit_one(self, x, y=None, **fit_params):
        """Fits each step with ``x``."""

        # Loop over the first n - 1 steps, which should all be transformers
        for t in itertools.islice(self.values(), len(self) - 1):
            x_pre = x
            x = t.transform_one(x=x)

            # If a transformer is supervised then it has to be updated
            if t.is_supervised:

                if isinstance(t, union.TransformerUnion):
                    for sub_t in t.values():
                        if sub_t.is_supervised:
                            sub_t.fit_one(x=x_pre, y=y)

                else:
                    t.fit_one(x=x_pre, y=y)

        self.final_estimator.fit_one(x=x, y=y, **fit_params)
        return self

    @metaestimators.if_delegate_has_method(delegate='final_estimator')
    def fit_predict_one(self, x, y, **fit_params):
        """Updates the pipeline and returns a the out-of-fold prediction.

        Only works if each estimator has a ``transform_one`` method and the final estimator has a
        ``fit_predict_one`` method.

        """
        x = self.transform_one(x=x)
        return self.final_estimator.fit_predict_one(x=x, y=y, **fit_params)

    @metaestimators.if_delegate_has_method(delegate='final_estimator')
    def fit_predict_proba_one(self, x, y, **fit_params):
        """Updates the pipeline and returns a the out-of-fold prediction.

        Only works if each estimator has a ``transform_one`` method and the final estimator has a
        ``fit_predict_one`` method.

        """
        x = self.transform_one(x=x)
        return self.final_estimator.fit_predict_proba_one(x=x, y=y, **fit_params)

    def transform_one(self, x):
        """Transform an input.

        Only works if each estimator has a ``transform_one`` method.

        """
        for transformer in self.transformers:

            if isinstance(transformer, union.TransformerUnion):

                # Fit the unsupervised part of the union
                for sub_transformer in transformer.values():
                    if not sub_transformer.is_supervised:
                        sub_transformer.fit_one(x=x)

            elif not transformer.is_supervised:
                transformer.fit_one(x=x)

            x = transformer.transform_one(x=x)

        return x

    @metaestimators.if_delegate_has_method(delegate='final_estimator')
    def predict_one(self, x):
        """Returns a prediction.

        Only works if each estimator has a ``transform_one`` method and the final estimator has a
        ``predict_one`` method.

        """
        x = self.transform_one(x=x)
        return self.final_estimator.predict_one(x=x)

    @metaestimators.if_delegate_has_method(delegate='final_estimator')
    def predict_proba_one(self, x):
        """Returns prediction probabilities.

        Only works if each estimator has a ``transform_one`` method and the final estimator has a
        ``predict_proba_one`` method.

        """
        x = self.transform_one(x=x)
        return self.final_estimator.predict_proba_one(x=x)

    @metaestimators.if_delegate_has_method(delegate='final_estimator')
    def forecast(self, horizon, xs=None):
        """Returns a forecast.

        Only works if each estimator has a ``transform_one`` method and the final estimator has a
        ``forecast`` method.

        """
        if xs is not None:
            xs = [self.transform_one(x) for x in xs]
        return self.final_estimator.forecast(horizon=horizon, xs=xs)

    def debug_one(self, x, show_types=True, n_decimals=5):
        """Displays the state of a set of features as it goes through the pipeline.

        Parameters:
            x (dict) A set of features.
            show_types (bool): Whether or not to display the type of feature along with it's value.
            n_decimals (int): Number of decimals to display for each floating point value.

        """

        TAB = ' ' * 4

        def format_value(x):
            if isinstance(x, float):
                return '{:.{prec}f}'.format(x, prec=n_decimals)
            return x

        def print_dict(x, show_types, indent=False, space_after=True):

            # Some transformers accept strings as input instead of dicts
            if isinstance(x, str):
                print(x)
            else:
                for k, v in sorted(x.items()):
                    type_str = f' ({type(v).__name__})' if show_types else ''
                    print((TAB if indent else '') + f'{k}: {format_value(v)}' + type_str)
            if space_after:
                print()

        def print_title(title, indent=False):
            print((TAB if indent else '') + title)
            print((TAB if indent else '') + '-' * len(title))

        # Print the initial state of the features
        print_title('0. Input')
        print_dict(x, show_types=show_types)

        # Print the state of x at each step
        for i, t in enumerate(self.transformers):

            if isinstance(t, union.TransformerUnion):
                print_title(f'{i+1}. Transformer union')
                for j, (name, sub_t) in enumerate(t.items()):
                    print_title(f'{i+1}.{j} {name}', indent=True)
                    print_dict(sub_t.transform_one(x), show_types=show_types, indent=True)
                x = t.transform_one(x)
                print_dict(x, show_types=show_types)

            else:
                print_title(f'{i+1}. {t}')
                x = t.transform_one(x)
                print_dict(x, show_types=show_types)

        # Print the predicted output from the final estimator
        final = self.final_estimator
        if not isinstance(final, base.Transformer):
            print_title(f'{len(self)}. {final}')

            if hasattr(final, 'debug_one'):
                final.debug_one(x)
                print()

            if isinstance(final, base.Classifier):
                print_dict(final.predict_proba_one(x), show_types=False, space_after=False)
            else:
                print(final.predict_one(x))

    def draw(self):
        """Draws the pipeline using the ``graphviz`` library."""

        def networkify(step):

            # Unions are converted to an undirected network
            if isinstance(step, union.TransformerUnion):
                return Network(nodes=map(networkify, step.values()), links=[], directed=False)

            # Pipelines are converted to a directed network
            if isinstance(step, Pipeline):
                return Network(
                    nodes=[],
                    links=zip(
                        map(networkify, list(step.values())[:-1]),
                        map(networkify, list(step.values())[1:])
                    ),
                    directed=True
                )

            # Wrapper models are handled recursively
            if isinstance(step, base.Wrapper):
                return Network(
                    nodes=[networkify(step._model)],
                    links=[],
                    directed=True,
                    name=type(step).__name__
                )

            # Other steps are treated as strings
            return str(step)

        # Draw input
        net = Network(nodes=['x'], links=[], directed=True)
        previous = 'x'

        # Draw each step
        for step in self.values():
            current = networkify(step)
            net.link(previous, current)
            previous = current

        # Draw output
        net.link(previous, 'y')

        return net.draw()



import math

from . import rmse


__all__ = ['RMSLE']


class RMSLE(rmse.RMSE):
    """Root mean squared logarithmic error.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [3, -0.5, 2, 7]
            >>> y_pred = [2.5, 0.0, 2, 8]

            >>> metric = metrics.RMSLE()
            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     print(metric.update(y_t, y_p).get())
            0.133531...
            0.499141...
            0.407546...
            0.357825...

            >>> metric
            RMSLE: 0.357826

    """

    def update(self, y_true, y_pred, sample_weight=1.):
        return super().update(math.log(y_true + 1), math.log(y_pred + 1), sample_weight)



import collections

from . import base


__all__ = ['Adam']


class Adam(base.Optimizer):
    """Adam optimizer.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_breast_cancer(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> optimizer = optim.Adam()
            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LogisticRegression(optimizer)
            ... )
            >>> metric = metrics.F1()

            >>> model_selection.online_score(X_y, model, metric)
            F1: 0.959554

    References:
        1. `Adam: A method for stochastic optimization <https://arxiv.org/pdf/1412.6980.pdf>`_

    """

    def __init__(self, lr=0.1, beta_1=0.9, beta_2=0.999, eps=1e-8):
        super().__init__(lr)
        self.beta_1 = beta_1
        self.beta_2 = beta_2
        self.eps = eps
        self.m = collections.defaultdict(float)
        self.v = collections.defaultdict(float)

    def _update_after_pred(self, w, g):

        for i, gi in g.items():
            self.m[i] = self.beta_1 * self.m[i] + (1 - self.beta_1) * gi
            self.v[i] = self.beta_2 * self.v[i] + (1 - self.beta_2) * gi ** 2
            m = self.m[i] / (1 - self.beta_1 ** (self.n_iterations + 1))
            v = self.v[i] / (1 - self.beta_2 ** (self.n_iterations + 1))
            w[i] -= self.learning_rate * m / (v ** 0.5 + self.eps)

        return w



from . import base
from . import ewmean


class EWVar(base.Univariate):
    """Exponentially weighted variance.

    To calculate the variance we use the fact that Var(X) = Mean(x^2) - Mean(x)^2 and internally
    we use the exponentially weighted mean of x/x^2 to calculate this.

    Parameters:
        alpha (float): The closer ``alpha`` is to 1 the more the statistic will adapt to recent
            values.

    Attributes:
        variance (float) : The running exponentially weighted variance.

    Example:

        ::

            >>> from creme import stats

            >>> X = [1, 3, 5, 4, 6, 8, 7, 9, 11]
            >>> ewv = stats.EWVar(alpha=0.5)
            >>> for x in X:
            ...     print(ewv.update(x).get())
            0
            1.0
            2.75
            1.4375
            1.984375
            3.43359375
            1.7958984375
            2.198974609375
            3.56536865234375

    References:
        1. `Incremental calculation of weighted mean and variance <http://people.ds.cam.ac.uk/fanf2/hermes/doc/antiforgery/stats.pdf>`_
        2. `Exponential Moving Average on Streaming Data <https://dev.to/nestedsoftware/exponential-moving-average-on-streaming-data-4hhl>`_
        3. `Pandas User Guide <http://pandas.pydata.org/pandas-docs/stable/user_guide/computation.html>`_

    """

    def __init__(self, alpha=0.5):
        self.alpha = alpha
        self.mean = ewmean.EWMean(alpha=alpha)
        self.sq_mean = ewmean.EWMean(alpha=alpha)

    @property
    def name(self):
        return f'ewv_{self.alpha}'

    def update(self, x):
        self.mean.update(x)
        self.sq_mean.update(x ** 2)
        return self

    def get(self):
        return self.sq_mean.get() - self.mean.get() ** 2



from .. import stats

from . import base


__all__ = ['Jaccard']


class Jaccard(stats.Mean, base.MultiOutputClassificationMetric):
    """Jaccard index for binary multi-outputs.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [
            ...     {0: False, 1: True, 2: True},
            ...     {0: True, 1: True, 2: False},
            ... ]

            >>> y_pred = [
            ...     {0: True, 1: True, 2: True},
            ...     {0: True, 1: False, 2: False},
            ... ]

            >>> jac = metrics.Jaccard()
            >>> for yt, yp in zip(y_true, y_pred):
            ...     jac = jac.update(yt, yp)

            >>> jac
            Jaccard: 0.583333

    References:
        1. `Similarity of asymmetric binary attributes <https://www.wikiwand.com/en/Jaccard_index#/Similarity_of_asymmetric_binary_attributes>`_

    """

    def bigger_is_better(self):
        return True

    def requires_labels(self):
        return True

    def update(self, y_true, y_pred):

        one_and_ones = 0
        zero_or_ones = 0

        for i in y_true:
            if y_true[i] and y_pred[i]:
                one_and_ones += 1
            elif y_true[i] or y_pred[i]:
                zero_or_ones += 1

        super().update(one_and_ones / (zero_or_ones + one_and_ones))

        return self



import collections
import copy

from .. import base
from .. import utils


__all__ = ['OneVsRestClassifier']


class OneVsRestClassifier(collections.UserDict, base.MultiClassifier):
    """One-vs-the-rest (OvR) multiclass strategy.

    This strategy consists in fitting one binary classifier per class. Because we are in a
    streaming context, the number of classes isn't known from the start, hence new classifiers are
    instantiated on the fly. Likewise the predicted probabilities will only include the classes
    seen up to a given point in time.

    Parameters:
        binary_classifier (base.BinaryClassifier)

    Attributes:
        classifiers (dict): A mapping between classes and classifiers.

    Example:

        ::

            >>> from creme import compose
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import multiclass
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_iris(),
            ...     shuffle=True,
            ...     random_state=42
            ... )

            >>> model = compose.Pipeline([
            ...     ('poly', preprocessing.PolynomialExtender(degree=2)),
            ...     ('scale', preprocessing.StandardScaler()),
            ...     ('learn', multiclass.OneVsRestClassifier(
            ...         binary_classifier=linear_model.LogisticRegression())
            ...     )
            ... ])

            >>> metric = metrics.MacroF1()

            >>> model_selection.online_score(X_y, model, metric)
            MacroF1: 0.801453

    """

    def __init__(self, binary_classifier: base.BinaryClassifier):
        super().__init__()
        self.binary_classifier = binary_classifier

    def __str__(self):
        return f'OneVsRestClassifier({self.binary_classifier})'

    def fit_one(self, x, y):

        # Instantiate a new binary classifier if the class is new
        if y not in self:
            self[y] = copy.deepcopy(self.binary_classifier)

        # Train each label's associated classifier
        for label, model in self.items():
            model.fit_one(x, y == label)

        return self

    def predict_proba_one(self, x):
        y_pred = {
            label: model.predict_proba_one(x)[True]
            for label, model in self.items()
        }
        return utils.math.softmax(y_pred)



"""
Utilities for compatibility with other libraries such as scikit-learn.
"""
import copy
import functools

import numpy as np
try:
    import pandas as pd
    PANDAS_INSTALLED = True
except ImportError:
    PANDAS_INSTALLED = False
from sklearn import base as sklearn_base
from sklearn import exceptions
from sklearn import pipeline
from sklearn import preprocessing
from sklearn import utils

from .. import base
from .. import compose
from .. import stream


__all__ = [
    'convert_creme_to_sklearn',
    'convert_sklearn_to_creme',
    'CremeClassifierWrapper',
    'CremeRegressorWrapper',
    'SKLRegressorWrapper',
    'SKLClassifierWrapper',
    'SKLClustererWrapper',
    'SKLTransformerWrapper'
]


STREAM_METHODS = {
    np.ndarray: stream.iter_array
}

if PANDAS_INSTALLED:
    STREAM_METHODS[pd.DataFrame] = stream.iter_pandas

# Params passed to sklearn.utils.check_X_y and sklearn.utils.check_array
SKLEARN_INPUT_X_PARAMS = {
    'accept_sparse': False,
    'accept_large_sparse': True,
    'dtype': 'numeric',
    'order': None,
    'copy': False,
    'force_all_finite': True,
    'ensure_2d': True,
    'allow_nd': False,
    'ensure_min_samples': 1,
    'ensure_min_features': 1
}

# Params passed to sklearn.utils.check_X_y in addition to SKLEARN_INPUT_X_PARAMS
SKLEARN_INPUT_Y_PARAMS = {
    'multi_output': False,
    'y_numeric': False
}


def convert_creme_to_sklearn(estimator):
    """Wraps a creme estimator to make it compatible with scikit-learn."""

    if isinstance(estimator, compose.Pipeline):
        return pipeline.Pipeline([
            (name, convert_creme_to_sklearn(step))
            for name, step in estimator.items()
        ])

    wrappers = [
        (base.BinaryClassifier, SKLClassifierWrapper),
        (base.Clusterer, SKLClustererWrapper),
        (base.MultiClassifier, SKLClassifierWrapper),
        (base.Regressor, SKLRegressorWrapper),
        (base.Transformer, SKLTransformerWrapper)
    ]

    for base_type, wrapper in wrappers:
        if isinstance(estimator, base_type):
            return wrapper(copy.deepcopy(estimator))

    raise ValueError("Couldn't find an appropriate wrapper")


def convert_sklearn_to_creme(estimator, classes=None):
    """Wraps an scikit-learn estimator to make it compatible with creme.

    Parameters:
        estimator (sklearn.base.BaseEstimator)
        classes (list): Class names necessary for classifiers.

    """

    wrappers = [
        (sklearn_base.RegressorMixin, CremeRegressorWrapper),
        (sklearn_base.ClassifierMixin, functools.partial(
            CremeClassifierWrapper,
            classes=classes
        ))
    ]

    for base_type, wrapper in wrappers:
        if isinstance(estimator, base_type):
            return wrapper(copy.deepcopy(estimator))

    raise ValueError("Couldn't find an appropriate wrapper")


class CremeBaseWrapper:

    def __init__(self, sklearn_estimator):
        self.sklearn_estimator = sklearn_estimator


class CremeRegressorWrapper(CremeBaseWrapper, base.Regressor):
    """Wraps an ``sklearn`` regressor to make it compatible with ``creme``.

    Example:

        ::

            >>> from creme import compat
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import linear_model
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_boston(),
            ...     shuffle=True,
            ...     random_state=42
            ... )

            >>> scaler = preprocessing.StandardScaler()
            >>> sgd_reg = compat.convert_sklearn_to_creme(linear_model.SGDRegressor())
            >>> model = scaler | sgd_reg

            >>> metric = metrics.MAE()

            >>> model_selection.online_score(X_y, model, metric)
            MAE: 10.832054

    """

    def fit_one(self, x, y):
        self.sklearn_estimator.partial_fit([list(x.values())], [y])
        return self

    def predict_one(self, x):
        try:
            return self.sklearn_estimator.predict([list(x.values())])[0]
        except exceptions.NotFittedError:
            return 0


class CremeClassifierWrapper(CremeBaseWrapper, base.MultiClassifier):
    """Wraps an ``sklearn`` classifier to make it compatible with ``creme``.

    Example:

        ::

            >>> from creme import compat
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import linear_model
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_breast_cancer(),
            ...     shuffle=True,
            ...     random_state=42
            ... )

            >>> model = preprocessing.StandardScaler()
            >>> model |= compat.convert_sklearn_to_creme(
            ...     estimator=linear_model.SGDClassifier(
            ...         loss='log',
            ...         eta0=0.01,
            ...         learning_rate='constant'
            ...     ),
            ...     classes=[False, True]
            ... )

            >>> metric = metrics.LogLoss()

            >>> model_selection.online_score(X_y, model, metric)
            LogLoss: 0.203717

    """

    def __init__(self, sklearn_estimator, classes):
        super().__init__(sklearn_estimator)
        self.classes = classes

    def fit_one(self, x, y):
        self.sklearn_estimator = self.sklearn_estimator.partial_fit(
            X=[list(x.values())],
            y=[y],
            classes=self.classes
        )
        return self

    def predict_proba_one(self, x):
        try:
            y_pred = self.sklearn_estimator.predict_proba([list(x.values())])[0]
            return {self.classes[i]: p for i, p in enumerate(y_pred)}
        except exceptions.NotFittedError:
            return {c: 1 / len(self.classes) for c in self.classes}

    def predict_one(self, x):
        try:
            y_pred = self.sklearn_estimator.predict([list(x.values())])[0]
            return y_pred
        except exceptions.NotFittedError:
            return None


class SKLBaseWrapper(sklearn_base.BaseEstimator):
    """The purpose of this class is to adapt the Sphinx documentation rendering."""

    def __init__(self, creme_estimator):
        self.creme_estimator = creme_estimator

    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Parameters:
            deep (bool, optional) If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns:
            dict of string to any: Parameter names mapped to their values.

        """
        return super().get_params(deep=deep)

    def set_params(self, **params):
        """Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects (such as pipelines). The
        latter have parameters of the form ``<component>__<parameter>`` so that it's possible to
        update each component of a nested object.

        Returns:
            self

        """
        return super().set_params(**params)


class SKLRegressorWrapper(SKLBaseWrapper, sklearn_base.RegressorMixin):

    def fit(self, X, y):
        """Fits to an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))
            y (array-like of shape n_samples)

        Returns:
            self

        """

        # Check the estimator is a Regressor
        if not isinstance(self.creme_estimator, base.Regressor):
            raise ValueError('creme_estimator is not a Regressor')

        # Check the inputs
        X, y = utils.check_X_y(X, y, **SKLEARN_INPUT_X_PARAMS, **SKLEARN_INPUT_Y_PARAMS)

        # scikit-learn's convention is that fit shouldn't mutate the input parameters; we have to
        # deep copy the provided estimator in order to respect this convention
        self.instance_ = copy.deepcopy(self.creme_estimator)

        # Call fit_one for each observation
        for x, yi in STREAM_METHODS[type(X)](X, y):
            self.instance_.fit_one(x, yi)

        # predict needs some way to know if fit has been called
        self.is_fitted_ = True

        return self

    def predict(self, X):
        """Predicts the target of an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))

        Returns:
            array of shape (n_samples,): Predicted target values for each row of ``X``.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, 'is_fitted_')

        # Check the input
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        # Make a prediction for each observation
        y_pred = np.empty(shape=len(X))
        for i, (x, _) in enumerate(stream.iter_array(X)):
            y_pred[i] = self.instance_.predict_one(x)

        return y_pred

    def score(self, X, y, sample_weight=None):
        """Returns the coefficient of determination R^2 of the prediction.

        The coefficient R^2 is defined as (1 - u/v), where u is the residual sum of squares
        ((y_true - y_pred) ** 2).sum() and v is the total sum of squares
        ((y_true - y_true.mean()) ** 2).sum(). The best possible score is 1.0 and it can be
        negative (because the model can be arbitrarily worse). A constant model that always
        predicts the expected value of y, disregarding the input features, would get a R^2 score of
        0.0.

        Parameters:

            X (array-like of shape (n_samples, n_features)): Test samples.
            y (array-like of shape = (n_samples) or (n_samples, n_outputs): True values for X.
            sample_weight (array-like of shape = [n_samples], optional): Sample weights.

        Returns:
            float: R^2 of self.predict(X) w.r.t. y.

        """
        return super().score(X, y, sample_weight)

    def fit_predict(self, X, y):
        """Calls ``fit`` and ``predict`` one after the other."""
        return self.fit(X, y).predict(X)


class SKLClassifierWrapper(SKLBaseWrapper, sklearn_base.ClassifierMixin):

    def fit(self, X, y):
        """Fits to an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))
            y (array-like of shape n_samples)

        Returns:
            self

        """

        # Check the estimator is either a BinaryClassifier or a MultiClassifier
        if not isinstance(self.creme_estimator, (base.BinaryClassifier, base.MultiClassifier)):
            raise ValueError('creme_estimator is not a BinaryClassifier nor a MultiClassifier')

        # Check the inputs
        X, y = utils.check_X_y(X, y, **SKLEARN_INPUT_X_PARAMS, **SKLEARN_INPUT_Y_PARAMS)

        # Check the target
        utils.multiclass.check_classification_targets(y)

        # Check the number of classes agrees with the type of classifier
        self.classes_ = np.unique(y)
        if len(self.classes_) > 2 and not isinstance(self.creme_estimator, base.MultiClassifier):
            raise ValueError(f'n_classes is more than 2 but {self.creme_estimator} is a ' +
                             'BinaryClassifier')

        # creme's BinaryClassifier expects bools or 0/1 values
        if not isinstance(self.creme_estimator, base.MultiClassifier):
            self.label_encoder_ = preprocessing.LabelEncoder().fit(y)
            y = self.label_encoder_.transform(y)

        # scikit-learn's convention is that fit shouldn't mutate the input parameters; we have to
        # deep copy the provided estimator in order to respect this convention
        self.instance_ = copy.deepcopy(self.creme_estimator)

        # Call fit_one for each observation
        for x, yi in STREAM_METHODS[type(X)](X, y):
            self.instance_.fit_one(x, yi)

        # Store the number of features so that future inputs can be checked
        self.n_features_ = X.shape[1]

        return self

    def predict_proba(self, X):
        """Predicts the target probability of an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))

        Returns:
            array of shape (n_samples,): Predicted target values for each row of ``X``.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, 'classes_')

        # Check the input
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        if X.shape[1] != self.n_features_:
            raise ValueError(f'Expected {self.n_features_} features, got {X.shape[1]}')

        # creme's predictions have to converted to follow the scikit-learn conventions
        def reshape_probas(y_pred):
            return [y_pred.get(c, 0) for c in self.classes_]

        # Make a prediction for each observation
        y_pred = np.empty(shape=(len(X), len(self.classes_)))
        for i, (x, _) in enumerate(stream.iter_array(X)):
            y_pred[i] = reshape_probas(self.instance_.predict_proba_one(x))

        return y_pred

    def predict(self, X):
        """Predicts the target of an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))

        Returns:
            array of shape (n_samples,): Predicted target values for each row of ``X``.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, 'classes_')

        # Check the input
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        if X.shape[1] != self.n_features_:
            raise ValueError(f'Expected {self.n_features_} features, got {X.shape[1]}')

        # Make a prediction for each observation
        y_pred = [None] * len(X)
        for i, (x, _) in enumerate(stream.iter_array(X)):
            y_pred[i] = self.instance_.predict_one(x)

        return np.asarray(y_pred)

    def score(self, X, y, sample_weight=None):
        """Returns the mean accuracy on the given test data and labels.
        In multi-label classification, this is the subset accuracy
        which is a harsh metric since you require for each sample that
        each label set be correctly predicted.

        Parameters:

            X (array-like of shape (n_samples, n_features)): Test samples.
            y (array-like of shape = (n_samples) or (n_samples, n_outputs): True values for X.

        Returns:
            float: Mean accuracy of self.predict(X) w.r.t. y.

        """
        return super().score(X, y, sample_weight)

    def fit_predict(self, X, y):
        """Calls ``fit`` and ``predict`` one after the other."""
        return self.fit(X, y).predict(X)


class SKLTransformerWrapper(SKLBaseWrapper, sklearn_base.TransformerMixin):

    def fit(self, X, y=None):
        """Fits to an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))
            y (array-like of shape n_samples)

        Returns:
            self

        """

        # Check the estimator is a Transformer
        if not isinstance(self.creme_estimator, base.Transformer):
            raise ValueError('creme_estimator is not a Transformer')

        # Check the inputs
        if y is None:
            X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)
        else:
            X, y = utils.check_X_y(X, y, **SKLEARN_INPUT_X_PARAMS, **SKLEARN_INPUT_Y_PARAMS)

        # scikit-learn's convention is that fit shouldn't mutate the input parameters; we have to
        # deep copy the provided estimator in order to respect this convention
        self.instance_ = copy.deepcopy(self.creme_estimator)

        # Call fit_one for each observation
        for x, yi in STREAM_METHODS[type(X)](X, y):
            self.instance_.fit_one(x, yi)

        # predict needs some way to know if fit has been called
        self.is_fitted_ = True

        # Store the number of features so that future inputs can be checked
        self.n_features_ = X.shape[1]

        return self

    def transform(self, X):
        """Predicts the target of an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))

        Returns:
            array: Transformed output.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, 'is_fitted_')

        # Check the input
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        if X.shape[1] != self.n_features_:
            raise ValueError(f'Expected {self.n_features_} features, got {X.shape[1]}')

        # Call predict_proba_one for each observation
        X_trans = [None] * len(X)
        for i, (x, _) in enumerate(stream.iter_array(X)):
            X_trans[i] = list(self.instance_.transform_one(x).values())

        return np.asarray(X_trans)

    def fit_transform(self, X, y=None):
        """Calls ``fit`` and ``transform`` one after the other."""
        return self.fit(X, y).transform(X)


class SKLClustererWrapper(SKLBaseWrapper, sklearn_base.ClusterMixin):

    def fit(self, X, y=None):
        """Fits to an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))
            y (array-like of shape n_samples)

        Returns:
            self

        """

        # Check the estimator is a Clusterer
        if not isinstance(self.creme_estimator, base.Clusterer):
            raise ValueError('creme_estimator is not a Clusterer')

        # Check the inputs
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        # scikit-learn's convention is that fit shouldn't mutate the input parameters; we have to
        # deep copy the provided estimator in order to respect this convention
        self.instance_ = copy.deepcopy(self.creme_estimator)

        # Call fit_one for each observation
        self.labels_ = np.empty(len(X), dtype=np.int32)
        for i, (x, yi) in enumerate(STREAM_METHODS[type(X)](X, y)):
            label = self.instance_.fit_one(x, yi).predict_one(x)
            self.labels_[i] = label

        # predict needs some way to know if fit has been called
        self.is_fitted_ = True

        return self

    def predict(self, X):
        """Predicts the target of an entire dataset contained in memory.

        Parameters:
            X (array-like of shape (n_samples, n_features))

        Returns:
            array: Transformed output.

        """

        # Check the fit method has been called
        utils.validation.check_is_fitted(self, 'is_fitted_')

        # Check the input
        X = utils.check_array(X, **SKLEARN_INPUT_X_PARAMS)

        # Call predict_proba_one for each observation
        y_pred = np.empty(len(X), dtype=np.int32)
        for i, (x, _) in enumerate(stream.iter_array(X)):
            y_pred[i] = self.instance_.predict_one(x)

        return y_pred

    def fit_predict(self, X, y=None):
        """Calls ``fit`` return the ``labels_`` computed attribute."""
        self.fit(X, y)
        return self.labels_



import collections
import copy
import functools

from .. import stats


__all__ = ['StatImputer']


class Constant(stats.Univariate):
    """Implements the ``stats.Univariate`` interface but always returns the same value.

    Arguments:
        value (any): A value

    """

    def __init__(self, value):
        self.value = value

    def update(self, x):
        return self

    def get(self):
        return self.value

    @property
    def name(self):
        return self.value


class StatImputer:
    """Imputer that allows to replace missing values with a univariate statistic, or a constant.

    Parameters:
        on (str): Name of the field to impute.
        by (str): Name of the field to impute with aggregatation.
        stat (stats.Univariate or any value): Univariate statistic used to fill missing values, for
            `stats.Mean`. If ``stat`` is not an instance of `stats.Univariate`, then the input will
            be treated as a constant and all missing values will be replaced by it.

    Example:

        ::

            >>> from creme import impute
            >>> from creme import stats

            >>> X = [
            ...     {'x': 1.0},
            ...     {'x': 2.0},
            ...     {'x': 3.0},
            ...     {}
            ... ]

            >>> const_imp = impute.StatImputer(
            ...     on='x',
            ...     stat=5.0
            ... )

            >>> for x in X:
            ...     print(const_imp.fit_one(x))
            {'x': 1.0}
            {'x': 2.0}
            {'x': 3.0}
            {'x': 5.0}

            >>> mean_imp = impute.StatImputer(
            ...     on='x',
            ...     stat=stats.Mean()
            ... )

            >>> for x in X:
            ...     print(mean_imp.fit_one(x))
            {'x': 1.0}
            {'x': 2.0}
            {'x': 3.0}
            {'x': 2.0}

            >>> X = [
            ...     {'x': 'sunny'},
            ...     {'x': 'rainy'},
            ...     {'x': 'humidity'},
            ...     {'x': 'sunny'},
            ...     {'x': 'rainy'},
            ...     {'x': 'rainy'},
            ...     {},
            ...     {},
            ...     {},
            ... ]

            >>> mode_imp = impute.StatImputer(
            ...     on='x',
            ...     stat=stats.Mode(),
            ... )

            >>> for x in X:
            ...     print(mode_imp.fit_one(x))
            {'x': 'sunny'}
            {'x': 'rainy'}
            {'x': 'humidity'}
            {'x': 'sunny'}
            {'x': 'rainy'}
            {'x': 'rainy'}
            {'x': 'rainy'}
            {'x': 'rainy'}
            {'x': 'rainy'}

            >>> X = [
            ...   {'town': 'New York', 'weather': 'sunny'},
            ...   {'town': 'New York', 'weather': 'sunny'},
            ...   {'town': 'New York', 'weather': 'rainy'},
            ...   {'town': 'Montreal', 'weather': 'rainy'},
            ...   {'town': 'Montreal', 'weather': 'humidity'},
            ...   {'town': 'Montreal', 'weather': 'rainy'},
            ...   {'town': 'Pekin', 'weather': 'sunny'},
            ...   {'town': 'New York'},
            ...   {'town': 'Montreal'},
            ...   {'town': 'Pekin'},
            ... ]

            >>> by_town_imp = impute.StatImputer(
            ...     on='weather',
            ...     by='town',
            ...     stat=stats.Mode()
            ... )

            >>> for x in X:
            ...     print(by_town_imp.fit_one(x))
            {'town': 'New York', 'weather': 'sunny'}
            {'town': 'New York', 'weather': 'sunny'}
            {'town': 'New York', 'weather': 'rainy'}
            {'town': 'Montreal', 'weather': 'rainy'}
            {'town': 'Montreal', 'weather': 'humidity'}
            {'town': 'Montreal', 'weather': 'rainy'}
            {'town': 'Pekin', 'weather': 'sunny'}
            {'town': 'New York', 'weather': 'sunny'}
            {'town': 'Montreal', 'weather': 'rainy'}
            {'town': 'Pekin', 'weather': 'sunny'}

    """

    def __init__(self, on, stat, by=None):
        self.on = on
        self.by = by
        self.stat = stat if isinstance(stat, stats.Univariate) else Constant(stat)
        self.imputers = collections.defaultdict(functools.partial(copy.deepcopy, self.stat))

    def fit_one(self, x):
        if self.on in x:
            key = x[self.by] if self.by else None
            self.imputers[key].update(x[self.on])
            return x
        return self.transform_one(x)

    def transform_one(self, x):
        if self.on not in x:
            key = x[self.by] if self.by else None
            return {
                **x,
                self.on: self.imputers[key].get()
            }
        return x



from .. import base


__all__ = ['FuncTransformer']


class FuncTransformer(base.Transformer):
    """Transforms a set of features following a given callable.

    The provided function has to take as input a ``dict`` of features and produce a new ``dict`` of
    transformed features.

    Parameters:
        func (callable): a function transforming a dict of features into a new dict of features.

    Example:

        ::

            >>> import datetime as dt
            >>> from creme import compose

            >>> x = {'date': '2019-02-14', 'x': 42}

            >>> def parse_date(x):
            ...     x['date'] = dt.datetime.strptime(x['date'], '%Y-%m-%d')
            ...     x['is_weekend'] = x['date'].day in (5, 6)
            ...     return x

            >>> transformer = compose.FuncTransformer(parse_date)
            >>> transformer.transform_one(x)
            {'date': datetime.datetime(2019, 2, 14, 0, 0), 'x': 42, 'is_weekend': False}

    """

    def __init__(self, func):
        self.func = func

    def transform_one(self, x):
        return self.func(x)

    def __str__(self):
        return self.func.__name__



from .. import stats

from . import base


__all__ = ['SMAPE']


class SMAPE(stats.Mean, base.RegressionMetric):
    """Symmetric mean absolute percentage error.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [0.07533, 0.07533, 0.07533, 0.07533, 0.07533, 0.07533, 0.0672, 0.0672]
            >>> y_pred = [0.102, 0.107, 0.047, 0.1, 0.032, 0.047, 0.108, 0.089]

            >>> metric = metrics.SMAPE()
            >>> for y_t, y_p in zip(y_true, y_pred):
            ...     metric = metric.update(y_t, y_p)

            >>> metric
            SMAPE: 42.603066

    """

    def update(self, y_true, y_pred, sample_weight=1.):
        return super().update(
            x=2. * abs(y_true - y_pred) / (abs(y_true) + abs(y_pred)),
            w=sample_weight
        )

    def get(self):
        return 100 * super().get()



import collections
import copy

from .. import base


__all__ = ['ClassifierChain', 'RegressorChain']


class BaseChain(collections.OrderedDict):

    def __init__(self, model, order):
        super().__init__()
        for o in order:
            self[o] = copy.deepcopy(model)

    def fit_one(self, x, y):

        x = copy.copy(x)

        for o, clf in self.items():
            y_pred = clf.predict_one(x)
            clf.fit_one(x, y[o])
            x[o] = y_pred

        return self


class ClassifierChain(BaseChain, base.MultiOutputClassifier):
    """A multi-label model that arranges classifiers into a chain.

    Example:

        ::

            >>> from creme import feature_selection
            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import multioutput
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.fetch_openml('yeast', version=4),
            ...     shuffle=True,
            ...     random_state=42
            ... )

            >>> model = feature_selection.VarianceThreshold(threshold=0.01)
            >>> model |= preprocessing.StandardScaler()
            >>> model |= multioutput.ClassifierChain(
            ...     model=linear_model.LogisticRegression(),
            ...     order=list(range(14))
            ... )

            >>> metric = metrics.Jaccard()

            >>> for x, y in X_y:
            ...     # Convert y values to booleans
            ...     y = {i: yi == 'TRUE' for i, yi in y.items()}
            ...     y_pred = model.predict_one(x)
            ...     metric = metric.update(y, y_pred)
            ...     model = model.fit_one(x, y)

            >>> metric
            Jaccard: 0.449923

    """

    def predict_proba_one(self, x):

        x = copy.copy(x)
        y_pred = {}

        for o, clf in self.items():
            y_pred[o] = clf.predict_proba_one(x)
            x[o] = max(y_pred[o], key=y_pred[o].get)

        return y_pred


class RegressorChain(BaseChain, base.MultiOutputRegressor):
    """A multi-label model that arranges regressor into a chain.

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import multioutput
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_linnerud(),
            ...     shuffle=True,
            ...     random_state=42
            ... )

            >>> model = multioutput.RegressorChain(
            ...     model=(
            ...         preprocessing.StandardScaler() |
            ...         linear_model.LinearRegression(intercept_lr=0.3)
            ...     ),
            ...     order=[0, 1, 2]
            ... )

            >>> metric = metrics.RegressionMultiOutput(metrics.MAE())

            >>> model_selection.online_score(X_y, model, metric)
            MAE: 16.970895...

    """

    def predict_one(self, x):

        x = copy.copy(x)
        y_pred = {}

        for o, clf in self.items():
            y_pred[o] = clf.predict_one(x)
            x[o] = y_pred[o]

        return y_pred



import collections

from . import base


class MiniBatcher(base.Optimizer):
    """Wrapper class for doing mini-batch gradient descent.

    This will accumulate successive gradients until ``batch_size`` have been observed. Once the
    current mini-batch is full, the underlying optimizer will update the weights using the average
    gradient. This can significantly improve the accuracy of a model, especially early on when
    gradients tend to go wild.

    Parameters:
        optimizer (optim.Optimizer)
        batch_size (int): The desired batch size.

    Attributes:
        gradient (collections.defaultdict): The current accumulated mini-batch.
        current_size (int): The current size of the mini-batch.
        n_iterations (int)

    Example:

        ::

            >>> from creme import linear_model
            >>> from creme import metrics
            >>> from creme import model_selection
            >>> from creme import optim
            >>> from creme import preprocessing
            >>> from creme import stream
            >>> from sklearn import datasets

            >>> X_y = stream.iter_sklearn_dataset(
            ...     dataset=datasets.load_breast_cancer(),
            ...     shuffle=True,
            ...     random_state=42
            ... )
            >>> optimizer = optim.MiniBatcher(optim.SGD(0.1), 4)
            >>> model = (
            ...     preprocessing.StandardScaler() |
            ...     linear_model.LogisticRegression(optimizer)
            ... )
            >>> metric = metrics.F1()

            >>> model_selection.online_score(X_y, model, metric)
            F1: 0.96756

    """

    def __init__(self, optimizer, batch_size):
        self.optimizer = optimizer
        self.batch_size = batch_size
        self.gradient = collections.defaultdict(float)
        self.current_size = 0
        self.n_iterations = 0

    def update_before_pred(self, w):
        return self.optimizer.update_before_pred(w)

    def _update_after_pred(self, w, g):

        # Accumulate the gradient
        for i, gi in g.items():
            self.gradient[i] += gi
        self.current_size += 1

        if self.current_size == self.batch_size:

            # Update the weights with the average gradient
            avg_g = {i: gi / self.batch_size for i, gi in self.gradient.items()}
            w = self.optimizer.update_after_pred(w, avg_g)

            # Reset the gradient
            self.gradient = collections.defaultdict(float)
            self.current_size = 0

        return w



from . import var


class SEM(var.Var):
    """Running standard error of the mean using Welford's algorithm.

    Parameters:
        ddof (int): Delta Degrees of Freedom. The divisor used in calculations is $n$ - ddof,
            where $n$ represents the number of seen elements.

    Attributes:
        n (int): Number of observations.

    Example:

        ::

            >>> import creme.stats

            >>> X = [3, 5, 4, 7, 10, 12]

            >>> sem = creme.stats.SEM()
            >>> for x in X:
            ...     print(sem.update(x).get())
            0.0
            1.0
            0.577350...
            0.853912...
            1.240967...
            1.447219...

    References:

        1. `Welford's online algorithm <https://www.wikiwand.com/en/Algorithms_for_calculating_variance#/Welford's_Online_algorithm>`_

    """

    def get(self):
        return (super().get() / self.mean.n) ** 0.5


class RollingSEM(var.RollingVar):
    """Running standard error of the mean over a window.

    Parameters:
        window_size (int): Size of the rolling window.
        ddof (int): Delta Degrees of Freedom for the variance.

    Example:

        ::
            >>> import creme

            >>> X = [1, 4, 2, -4, -8, 0]

            >>> rolling_sem = creme.stats.RollingSEM(ddof=1, window_size=2)
            >>> for x in X:
            ...     print(rolling_sem.update(x).get())
            0.0
            1.5
            1.0
            3.0
            2.0
            4.0

            >>> rolling_sem = creme.stats.RollingSEM(ddof=1, window_size=3)
            >>> for x in X:
            ...     print(rolling_sem.update(x).get())
            0.0
            1.5
            0.881917...
            2.403700...
            2.905932...
            2.309401...

    """

    def get(self):
        return (super().get() / len(self.rolling_mean)) ** 0.5



import collections

from .. import stats

from . import base


__all__ = [
    'MacroPrecision',
    'MicroPrecision',
    'Precision',
    'WeightedPrecision'
]


class BasePrecision:

    @property
    def bigger_is_better(self):
        return True

    @property
    def requires_labels(self):
        return True


class Precision(stats.Mean, BasePrecision, base.BinaryMetric):
    """Binary precision score.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [True, False, True, True, True]
            >>> y_pred = [True, True, False, True, True]

            >>> metric = metrics.Precision()

            >>> for yt, yp in zip(y_true, y_pred):
            ...     print(metric.update(yt, yp))
            Precision: 1.
            Precision: 0.5
            Precision: 0.5
            Precision: 0.666667
            Precision: 0.75

    """

    def update(self, y_true, y_pred, sample_weight=1.):
        if y_pred:
            super().update(x=y_true == y_pred, w=sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        if y_pred:
            super().revert(x=y_true == y_pred, w=sample_weight)
        return self


class MacroPrecision(BasePrecision, base.MultiClassMetric):
    """Macro-average precision score.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [0, 1, 2, 2, 2]
            >>> y_pred = [0, 0, 2, 2, 1]

            >>> metric = metrics.MacroPrecision()

            >>> for yt, yp in zip(y_true, y_pred):
            ...     print(metric.update(yt, yp))
            MacroPrecision: 1.
            MacroPrecision: 0.25
            MacroPrecision: 0.5
            MacroPrecision: 0.5
            MacroPrecision: 0.5

    """

    def __init__(self):
        self.precisions = collections.defaultdict(Precision)
        self._class_counts = collections.Counter()

    def update(self, y_true, y_pred, sample_weight=1.):
        self.precisions[y_pred].update(y_true == y_pred, True, sample_weight)
        self._class_counts.update([y_true, y_pred])
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.precisions[y_pred].revert(y_true == y_pred, True, sample_weight)
        self._class_counts.subtract([y_true, y_pred])
        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return sum(self.precisions[c].get() for c in relevant) / len(relevant)
        except ZeroDivisionError:
            return 0.


class MicroPrecision(stats.Mean, BasePrecision, base.MultiClassMetric):
    """Micro-average precision score.

    The micro-average precision score is exactly equivalent to the micro-average recall as well as
    the micro-average F1 score.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [0, 1, 2, 2, 2]
            >>> y_pred = [0, 0, 2, 2, 1]

            >>> metric = metrics.MicroPrecision()

            >>> for yt, yp in zip(y_true, y_pred):
            ...     print(metric.update(yt, yp))
            MicroPrecision: 1.
            MicroPrecision: 0.5
            MicroPrecision: 0.666667
            MicroPrecision: 0.75
            MicroPrecision: 0.6

    References:
        1. `Why are precision, recall and F1 score equal when using micro averaging in a multi-class problem? <https://simonhessner.de/why-are-precision-recall-and-f1-score-equal-when-using-micro-averaging-in-a-multi-class-problem/>`_

    """

    def update(self, y_true, y_pred, sample_weight=1.):
        super().update(x=y_true == y_pred, w=sample_weight)
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        super().revert(x=y_true == y_pred, w=sample_weight)
        return self


class WeightedPrecision(BasePrecision, base.MultiClassMetric):
    """Weighted-average precision score.

    This uses the support of each label to compute an average score, whereas `MacroPrecision`
    ignores the support.

    Example:

        ::

            >>> from creme import metrics

            >>> y_true = [0, 1, 2, 2, 2]
            >>> y_pred = [0, 0, 2, 2, 1]

            >>> metric = metrics.WeightedPrecision()

            >>> for yt, yp in zip(y_true, y_pred):
            ...     print(metric.update(yt, yp))
            WeightedPrecision: 1.
            WeightedPrecision: 0.25
            WeightedPrecision: 0.5
            WeightedPrecision: 0.625
            WeightedPrecision: 0.7

    """

    def __init__(self):
        self.precisions = collections.defaultdict(Precision)
        self.support = collections.Counter()
        self._class_counts = collections.Counter()

    def update(self, y_true, y_pred, sample_weight=1.):
        self.precisions[y_pred].update(y_true == y_pred, True, sample_weight)
        self.support.update({y_true: sample_weight})
        self._class_counts.update([y_true, y_pred])
        return self

    def revert(self, y_true, y_pred, sample_weight=1.):
        self.precisions[y_pred].revert(y_true == y_pred, True, sample_weight)
        self.support.subtract({y_true: sample_weight})
        self._class_counts.subtract([y_true, y_pred])
        return self

    def get(self):
        relevant = [c for c, count in self._class_counts.items() if count > 0]
        try:
            return (
                sum(self.precisions[c].get() * self.support[c] for c in relevant) /
                sum(self.support[c] for c in relevant)
            )
        except ZeroDivisionError:
            return 0.
